import itertools
import numpy
import pytest
import time

from helpers import *
from .cbrng_ref import philox as philox_ref
from .cbrng_ref import threefry as threefry_ref

from reikna.core import Type
from reikna.helpers import product
from reikna.cbrng import CBRNG
from reikna.cbrng.bijections import threefry, philox
from reikna.cbrng.tools import KeyGenerator
from reikna.cbrng.samplers import uniform_integer, uniform_float, normal_bm, gamma
import reikna.cluda.dtypes as dtypes


def uniform_discrete_mean_and_std(min, max):
    return (min + max) / 2., numpy.sqrt(((max - min + 1) ** 2 - 1.) / 12)


def uniform_mean_and_std(min, max):
    return (min + max) / 2., (max - min) / numpy.sqrt(12)


class TestUniformInteger:
    def __init__(self, bijection, min_, max_):
        self.extent = (min_, max_)
        self.mean, self.std = uniform_discrete_mean_and_std(*self.extent)
        self.name = 'uniform_integer'
        self.sampler = uniform_integer(bijection, numpy.int32, min_, max_ + 1)

class TestUniformFloat:
    def __init__(self, bijection, min_, max_):
        self.extent = (min_, max_)
        self.mean, self.std = uniform_mean_and_std(*self.extent)
        self.bijection = bijection
        self.name = 'uniform_float'

    def get_sampler(self, double):
        dtype = numpy.float64 if double else numpy.float32
        return uniform_float(self.bijection, dtype, self.extent[0], self.extent[1])

class TestNormalBM:
    def __init__(self, bijection, mean, std):
        self.extent = None
        self.mean = mean
        self.std = std
        self.bijection = bijection
        self.name = 'normal_bm'

    def get_sampler(self, double):
        dtype = numpy.float64 if double else numpy.float32
        return normal_bm(self.bijection, dtype, mean=self.mean, std=self.std)

class TestNormalBMComplex:
    def __init__(self, bijection, mean, std):
        self.extent = None
        self.mean = mean
        self.std = std
        self.bijection = bijection
        self.name = 'normal_bm_complex'

    def get_sampler(self, double):
        dtype = numpy.complex128 if double else numpy.complex64
        return normal_bm(self.bijection, dtype, mean=self.mean, std=self.std)

class TestGamma:
    def __init__(self, bijection, shape, scale):
        self._shape = shape
        self._scale = scale
        self.mean = shape * scale
        self.std = numpy.sqrt(shape) * scale
        self.extent = None
        self.bijection = bijection
        self.name = 'gamma'

    def get_sampler(self, double):
        dtype = numpy.float64 if double else numpy.float32
        return gamma(self.bijection, dtype, shape=self._shape, scale=self._scale)


class TestBijection:

    def __init__(self, name, words, bitness):
        rounds = 20 if name == 'threefry' else 10
        if name == 'philox':
            bijection_func = philox
        else:
            bijection_func = threefry
        self._name = name
        self._words = words
        self._bitness = bitness
        self._rounds = rounds
        self.bijection = bijection_func(bitness, words, rounds=rounds)

        func = philox_ref if name == 'philox' else threefry_ref
        self._reference_func = lambda ctr, key: func(bitness, words, ctr, key, Nrounds=rounds)

    def reference(self, counters, keygen):
        result = numpy.empty_like(counters)
        for i in range(counters.shape[0]):
            result[i] = self._reference_func(counters[i]['v'], keygen(i)['v'])
        return result

    def __str__(self):
        return "{name}-{words}x{bitness}-{rounds}".format(
            name=self._name, words=self._words, bitness=self._bitness, rounds=self._rounds)


def pytest_generate_tests(metafunc):

    if 'test_bijection' in metafunc.funcargnames:

        vals = []
        ids = []

        for name, words, bitness in itertools.product(['threefry', 'philox'], [2, 4], [32, 64]):
            val = TestBijection(name, words, bitness)
            vals.append(val)
            ids.append(str(val))

        metafunc.parametrize('test_bijection', vals, ids=ids)

    if 'test_sampler_int' in metafunc.funcargnames:
        bijection = philox(64, 4)
        vals = [
            TestUniformInteger(bijection, -10, 98)]
        ids = [test.name for test in vals]
        metafunc.parametrize('test_sampler_int', vals, ids=ids)

    if 'test_sampler_float' in metafunc.funcargnames:
        bijection = philox(64, 4)
        vals = [
            TestUniformFloat(bijection, -5, 7.7),
            TestNormalBM(bijection, -2, 10),
            TestNormalBMComplex(bijection, -3 + 4j, 7),
            TestGamma(bijection, 3, 10)]
        ids = [test.name for test in vals]
        metafunc.parametrize('test_sampler_float', vals, ids=ids)


def test_kernel_bijection(thr, test_bijection):

    size = 1000
    seed = 123

    bijection = test_bijection.bijection
    keygen = KeyGenerator.create(bijection, seed=seed, reserve_id_space=False)
    counters_ref = numpy.zeros(size, bijection.counter_dtype)

    rng_kernel = thr.compile_static(
        """
        KERNEL void test(GLOBAL_MEM ${bijection.module}Counter *dest, int ctr)
        {
            VIRTUAL_SKIP_THREADS;
            const VSIZE_T idx = virtual_global_id(0);

            ${bijection.module}Key key = ${keygen.module}key_from_int(idx);
            ${bijection.module}Counter counter = ${bijection.module}make_counter_from_int(ctr);
            ${bijection.module}Counter result = ${bijection.module}bijection(key, counter);

            dest[idx] = result;
        }
        """,
        'test', size,
        render_kwds=dict(bijection=bijection, keygen=keygen))

    dest = thr.array(size, bijection.counter_dtype)

    rng_kernel(dest, numpy.int32(0))
    dest_ref = test_bijection.reference(counters_ref, keygen.reference)
    assert (dest.get() == dest_ref).all()

    rng_kernel(dest, numpy.int32(1))
    counters_ref['v'][:,-1] = 1
    dest_ref = test_bijection.reference(counters_ref, keygen.reference)
    assert (dest.get() == dest_ref).all()


def check_kernel_sampler(thr, sampler, extent=None, mean=None, std=None):

    size = 10000
    batch = 100
    seed = 456

    bijection = sampler.bijection
    keygen = KeyGenerator.create(bijection, seed=seed)

    rng_kernel = thr.compile_static(
        """
        KERNEL void test(GLOBAL_MEM ${ctype} *dest, int ctr_start)
        {
            VIRTUAL_SKIP_THREADS;
            const VSIZE_T idx = virtual_global_id(0);

            ${bijection.module}Key key = ${keygen.module}key_from_int(idx);
            ${bijection.module}Counter ctr = ${bijection.module}make_counter_from_int(ctr_start);
            ${bijection.module}State st = ${bijection.module}make_state(key, ctr);

            ${sampler.module}Result res;

            for(int j = 0; j < ${batch}; j++)
            {
                res = ${sampler.module}sample(&st);

                %for i in range(sampler.randoms_per_call):
                dest[j * ${size * sampler.randoms_per_call} + ${size * i} + idx] = res.v[${i}];
                %endfor
            }

            ${bijection.module}Counter next_ctr = ${bijection.module}get_next_unused_counter(st);
        }
        """,
        'test', size,
        render_kwds=dict(
            size=size, batch=batch, ctype=dtypes.ctype(sampler.dtype),
            bijection=bijection, keygen=keygen, sampler=sampler))

    dest = thr.array((batch, sampler.randoms_per_call, size), sampler.dtype)
    rng_kernel(dest, numpy.int32(0))
    dest = dest.get()

    check_distribution(dest, extent=extent, mean=mean, std=std)


def check_distribution(arr, extent=None, mean=None, std=None):
    if extent is not None:
        assert arr.min() >= extent[0]
        assert arr.max() <= extent[1]

    if mean is not None and std is not None:
        # expected mean and std of the mean of the sample array
        m_mean = mean
        m_std = std / numpy.sqrt(arr.size)

        diff = abs(arr.mean() - mean)
        assert diff < 5 * m_std # about 1e-6 chance of fail

    if std is not None:
        # expected mean and std of the variance of the sample array
        v_mean = std ** 2
        v_std = numpy.sqrt(2. * std ** 4 / (arr.size - 1))

        diff = abs(arr.var() - v_mean)
        assert diff < 5 * v_std # about 1e-6 chance of fail


def test_32_to_64_bit(thr):
    extent = (0, 2**63-1)
    mean, std = uniform_discrete_mean_and_std(*extent)
    bijection = philox(32, 4)
    sampler = uniform_integer(bijection, numpy.uint64, extent[0], extent[1] + 1)
    check_kernel_sampler(thr, sampler, extent=extent, mean=mean, std=std)


def test_64_to_32_bit(thr):
    extent = (0, 2**31-1)
    mean, std = uniform_discrete_mean_and_std(*extent)
    bijection = philox(64, 4)
    sampler = uniform_integer(bijection, numpy.uint32, extent[0], extent[1] + 1)
    check_kernel_sampler(thr, sampler, extent=extent, mean=mean, std=std)


def test_kernel_sampler_int(thr, test_sampler_int):
    check_kernel_sampler(thr, test_sampler_int.sampler,
        extent=test_sampler_int.extent, mean=test_sampler_int.mean, std=test_sampler_int.std)

def test_kernel_sampler_float(thr_and_double, test_sampler_float):
    thr, double = thr_and_double
    check_kernel_sampler(thr, test_sampler_float.get_sampler(double),
        extent=test_sampler_float.extent,
        mean=test_sampler_float.mean, std=test_sampler_float.std)


def check_computation(thr, rng, extent=None, mean=None, std=None):
    dest_dev = thr.empty_like(rng.parameter.randoms)
    counters = rng.create_counters()
    counters_dev = thr.to_device(counters)
    rngc = rng.compile(thr)

    rngc(counters_dev, dest_dev)
    dest = dest_dev.get()
    check_distribution(dest, extent=extent, mean=mean, std=std)


def test_computation_general(thr_and_double):

    size = 10000
    batch = 101

    thr, double = thr_and_double
    dtype = numpy.float64 if double else numpy.float32
    mean, std = -2, 10
    bijection = philox(64, 4)
    sampler = normal_bm(bijection, dtype, mean=mean, std=std)

    rng = CBRNG(Type(dtype, shape=(batch, size)), 1, sampler)
    check_computation(thr, rng, mean=mean, std=std)


def test_computation_convenience(thr):

    size = 10000
    batch = 101

    extent = (0, 511)
    mean, std = uniform_discrete_mean_and_std(*extent)
    rng = CBRNG.uniform_integer(Type(numpy.int32, shape=(batch, size)), 1,
        sampler_kwds=dict(low=extent[0], high=extent[1] + 1))
    check_computation(thr, rng, extent=extent, mean=mean, std=std)


def test_computation_uniqueness(thr):
    """
    A regression test for the bug with a non-updating counter.
    """

    size = 10000
    batch = 1

    rng = CBRNG.normal_bm(Type(numpy.complex64, shape=(batch, size)), 1)

    dest1_dev = thr.empty_like(rng.parameter.randoms)
    dest2_dev = thr.empty_like(rng.parameter.randoms)
    counters = rng.create_counters()
    counters_dev = thr.to_device(counters)
    rngc = rng.compile(thr)

    rngc(counters_dev, dest1_dev)
    rngc(counters_dev, dest2_dev)

    assert not diff_is_negligible(dest1_dev.get(), dest2_dev.get())


@pytest.mark.perf
@pytest.mark.returns('GB/s')
def test_computation_performance(thr_and_double, fast_math, test_sampler_float):

    thr, double = thr_and_double

    size = 2 ** 15
    batch = 2 ** 6

    sampler = test_sampler_float.get_sampler(double)

    rng = CBRNG(Type(sampler.dtype, shape=(batch, size)), 1, sampler)

    dest_dev = thr.empty_like(rng.parameter.randoms)
    counters = rng.create_counters()
    counters_dev = thr.to_device(counters)
    rngc = rng.compile(thr, fast_math=fast_math)

    attempts = 10
    times = []
    for i in range(attempts):
        t1 = time.time()
        rngc(counters_dev, dest_dev)
        thr.synchronize()
        times.append(time.time() - t1)

    byte_size = size * batch * sampler.dtype.itemsize
    return min(times), byte_size



import numpy
from numpy.lib.stride_tricks import as_strided

from reikna.helpers import wrap_in_tuple
from reikna.cluda import dtypes

SINGLE_EPS = 1e-6
DOUBLE_EPS = 1e-11


def get_test_array_like(arr, **kwds):
    kwds['strides'] = arr.strides
    return get_test_array(arr.shape, arr.dtype, **kwds)

def get_test_array(shape, dtype, strides=None, no_zeros=False, high=None):
    shape = wrap_in_tuple(shape)
    dtype = dtypes.normalize_type(dtype)

    if dtype.names is not None:
        result = numpy.empty(shape, dtype)
        for name in dtype.names:
            result[name] = get_test_array(shape, dtype[name], no_zeros=no_zeros, high=high)
    else:
        if dtypes.is_integer(dtype):
            low = 1 if no_zeros else 0
            if high is None:
                high = 100 # will work even with signed chars
            get_arr = lambda: numpy.random.randint(low, high, shape).astype(dtype)
        else:
            low = 0.01 if no_zeros else 0
            if high is None:
                high = 1.0
            get_arr = lambda: numpy.random.uniform(low, high, shape).astype(dtype)

        if dtypes.is_complex(dtype):
            result = get_arr() + 1j * get_arr()
        else:
            result = get_arr()

    if strides is not None:
        result = as_strided(result, result.shape, strides)

    return result

def float_diff(m, m_ref):
    return numpy.linalg.norm(m - m_ref) / numpy.linalg.norm(m_ref)

def diff_is_negligible(m, m_ref):

    if m.dtype.names is not None:
        return all(diff_is_negligible(m[name], m_ref[name]) for name in m.dtype.names)

    assert m.dtype == m_ref.dtype

    if dtypes.is_integer(m.dtype):
        return ((m - m_ref) == 0).all()

    diff = float_diff(m, m_ref)
    if dtypes.is_double(m.dtype):
        return diff < DOUBLE_EPS
    else:
        return diff < SINGLE_EPS



import numpy
import pytest

from reikna.core import Parameter, Annotation, Transformation
from reikna.core.signature import Type
from reikna.cluda import functions

from helpers import *
from test_core.dummy import *


# Some transformations to use by tests

# Identity transformation: Output = Input
def tr_identity(arr):
    return Transformation(
        [Parameter('o1', Annotation(arr, 'o')),
        Parameter('i1', Annotation(arr, 'i'))],
        "${o1.store_same}(${i1.load_same});")

# Output = Input1 * Parameter1 + Input 2
def tr_2_to_1(arr, scalar):
    return Transformation(
        [Parameter('o1', Annotation(arr, 'o')),
        Parameter('i1', Annotation(arr, 'i')),
        Parameter('i2', Annotation(arr, 'i')),
        Parameter('s1', Annotation(scalar))],
        """
        ${o1.ctype} t = ${mul}(${cast}(${s1}), ${i1.load_same});
        ${o1.store_same}(t + ${i2.load_same});
        """,
        render_kwds= dict(
            mul=functions.mul(arr.dtype, arr.dtype),
            cast=functions.cast(arr.dtype, scalar.dtype)))

# Output1 = Input / 2, Output2 = Input / 2
def tr_1_to_2(arr):
    return Transformation(
        [Parameter('o1', Annotation(arr, 'o')),
        Parameter('o2', Annotation(arr, 'o')),
        Parameter('i1', Annotation(arr, 'i'))],
        """
        ${o1.ctype} t = ${mul}(${i1.load_same}, 0.5);
        ${o1.store_same}(t);
        ${o2.store_same}(t);
        """,
        render_kwds=dict(
            mul=functions.mul(arr.dtype, numpy.float32)))


def test_io_parameter_in_transformation():
    with pytest.raises(ValueError):
        tr = Transformation(
            [Parameter('o1', Annotation(Type(numpy.float32, shape=100), 'io'))],
            "${o1.store_same}(${o1.load_same});")


def test_signature_correctness():

    N = 200
    coeff_dtype = numpy.float32
    arr_type = Type(numpy.complex64, (N, N))

    d = Dummy(arr_type, arr_type, coeff_dtype, same_A_B=True)

    # Root signature
    assert list(d.signature.parameters.values()) == [
        Parameter('C', Annotation(arr_type, 'o')),
        Parameter('D', Annotation(arr_type, 'o')),
        Parameter('A', Annotation(arr_type, 'i')),
        Parameter('B', Annotation(arr_type, 'i')),
        Parameter('coeff', Annotation(coeff_dtype))]

    identity = tr_identity(d.parameter.A)
    join = tr_2_to_1(d.parameter.A, d.parameter.coeff)
    split = tr_1_to_2(d.parameter.A)
    scale = tr_scale(d.parameter.A, d.parameter.coeff.dtype)

    # Connect some transformations
    d.parameter.A.connect(identity, identity.o1, A_prime=identity.i1)
    d.parameter.B.connect(join, join.o1, A_prime=join.i1, B_prime=join.i2, B_param=join.s1)
    d.parameter.B_prime.connect(identity, identity.o1, B_new_prime=identity.i1)
    d.parameter.C.connect(split, split.i1, C_half1=split.o1, C_half2=split.o2)
    d.parameter.C_half1.connect(identity, identity.i1, C_new_half1=identity.o1)
    d.parameter.D.connect(scale, scale.i1, D_prime=scale.o1, D_param=scale.s1)

    assert list(d.signature.parameters.values()) == [
        Parameter('C_new_half1', Annotation(arr_type, 'o')),
        Parameter('C_half2', Annotation(arr_type, 'o')),
        Parameter('D_prime', Annotation(arr_type, 'o')),
        Parameter('D_param', Annotation(coeff_dtype)),
        Parameter('A_prime', Annotation(arr_type, 'i')),
        Parameter('B_new_prime', Annotation(arr_type, 'i')),
        Parameter('B_param', Annotation(coeff_dtype)),
        Parameter('coeff', Annotation(coeff_dtype))]


def test_same_shape(thr):

    N = 200
    coeff = 2
    B_param = 3
    D_param = 4

    coeff_dtype = numpy.float32
    arr_type = Type(numpy.complex64, (N, N))

    d = Dummy(arr_type, arr_type, coeff_dtype, same_A_B=True)

    identity = tr_identity(d.parameter.A)
    join = tr_2_to_1(d.parameter.A, d.parameter.coeff)
    split = tr_1_to_2(d.parameter.A)
    scale = tr_scale(d.parameter.A, d.parameter.coeff.dtype)

    d.parameter.A.connect(identity, identity.o1, A_prime=identity.i1)
    d.parameter.B.connect(join, join.o1, A_prime=join.i1, B_prime=join.i2, B_param=join.s1)
    d.parameter.B_prime.connect(identity, identity.o1, B_new_prime=identity.i1)
    d.parameter.C.connect(split, split.i1, C_half1=split.o1, C_half2=split.o2)
    d.parameter.C_half1.connect(identity, identity.i1, C_new_half1=identity.o1)
    d.parameter.D.connect(scale, scale.i1, D_prime=scale.o1, D_param=scale.s1)
    dc = d.compile(thr)

    A_prime = get_test_array_like(d.parameter.A_prime)
    B_new_prime = get_test_array_like(d.parameter.B_new_prime)

    A_prime_dev = thr.to_device(A_prime)
    B_new_prime_dev = thr.to_device(B_new_prime)
    C_new_half1_dev = thr.empty_like(d.parameter.A_prime)
    C_half2_dev = thr.empty_like(d.parameter.A_prime)
    D_prime_dev = thr.empty_like(d.parameter.A_prime)

    dc(
        C_new_half1_dev, C_half2_dev,
        D_prime_dev, D_param,
        A_prime_dev, B_new_prime_dev,
        B_param, coeff)

    A = A_prime
    B = A_prime * B_param + B_new_prime
    C, D = mock_dummy(A, B, coeff)
    C_new_half1 = C / 2
    C_half2 = C / 2
    D_prime = D * D_param

    assert diff_is_negligible(C_new_half1_dev.get(), C_new_half1)
    assert diff_is_negligible(C_half2_dev.get(), C_half2)
    assert diff_is_negligible(D_prime_dev.get(), D_prime)


def test_connection_to_base(thr):

    N = 200
    coeff = 2

    coeff_dtype = numpy.float32
    arr_type = Type(numpy.complex64, (N, N))

    d = Dummy(arr_type, arr_type, coeff_dtype, same_A_B=True)

    identity = tr_identity(d.parameter.A)
    scale = tr_scale(d.parameter.A, d.parameter.coeff.dtype)

    # connect to the base array argument (effectively making B the same as A)
    d.parameter.A.connect(identity, identity.o1, B=identity.i1)
    # connect to the base scalar argument
    d.parameter.C.connect(scale, scale.i1, C_prime=scale.o1, coeff=scale.s1)

    assert list(d.signature.parameters.values()) == [
        Parameter('C_prime', Annotation(arr_type, 'o')),
        Parameter('D', Annotation(arr_type, 'o')),
        Parameter('B', Annotation(arr_type, 'i')),
        Parameter('coeff', Annotation(coeff_dtype))]

    dc = d.compile(thr)

    B = get_test_array_like(d.parameter.B)
    B_dev = thr.to_device(B)
    C_prime_dev = thr.empty_like(d.parameter.B)
    D_dev = thr.empty_like(d.parameter.B)

    dc(C_prime_dev, D_dev, B_dev, coeff)

    C, D = mock_dummy(B, B, coeff)
    C_prime = C * coeff

    assert diff_is_negligible(C_prime_dev.get(), C_prime)
    assert diff_is_negligible(D_dev.get(), D)


def test_nested_same_shape(thr):

    N = 2000
    coeff = 2
    second_coeff = 7
    B_param = 3
    D_param = 4

    coeff_dtype = numpy.float32
    arr_type = Type(numpy.complex64, (N, N))

    d = DummyNested(arr_type, arr_type, coeff_dtype, second_coeff, same_A_B=True)

    identity = tr_identity(d.parameter.A)
    join = tr_2_to_1(d.parameter.A, d.parameter.coeff)
    split = tr_1_to_2(d.parameter.A)
    scale = tr_scale(d.parameter.A, d.parameter.coeff.dtype)

    d.parameter.A.connect(identity, identity.o1, A_prime=identity.i1)
    d.parameter.B.connect(join, join.o1, A_prime=join.i1, B_prime=join.i2, B_param=join.s1)
    d.parameter.B_prime.connect(identity, identity.o1, B_new_prime=identity.i1)
    d.parameter.C.connect(split, split.i1, C_half1=split.o1, C_half2=split.o2)
    d.parameter.C_half1.connect(identity, identity.i1, C_new_half1=identity.o1)
    d.parameter.D.connect(scale, scale.i1, D_prime=scale.o1, D_param=scale.s1)
    dc = d.compile(thr)

    A_prime = get_test_array_like(d.parameter.A_prime)
    B_new_prime = get_test_array_like(d.parameter.B_new_prime)

    A_prime_dev = thr.to_device(A_prime)
    B_new_prime_dev = thr.to_device(B_new_prime)
    C_new_half1_dev = thr.empty_like(d.parameter.A_prime)
    C_half2_dev = thr.empty_like(d.parameter.A_prime)
    D_prime_dev = thr.empty_like(d.parameter.A_prime)

    dc(
        C_new_half1_dev, C_half2_dev,
        D_prime_dev, D_param,
        A_prime_dev, B_new_prime_dev,
        B_param, coeff)

    A = A_prime
    B = A_prime * B_param + B_new_prime
    C, D = mock_dummy_nested(A, B, coeff, second_coeff)
    C_new_half1 = C / 2
    C_half2 = C / 2
    D_prime = D * D_param

    assert diff_is_negligible(C_new_half1_dev.get(), C_new_half1)
    assert diff_is_negligible(C_half2_dev.get(), C_half2)
    assert diff_is_negligible(D_prime_dev.get(), D_prime)


def test_strings_as_parameters():
    """
    Check that one can connect transformations using strings as identifiers
    for computation and transformation parameters.
    """

    N = 200
    coeff_dtype = numpy.float32
    arr_type = Type(numpy.complex64, (N, N))

    d = Dummy(arr_type, arr_type, coeff_dtype, same_A_B=True)
    identity = tr_identity(d.parameter.A)

    d.connect('A', identity, 'o1', A_prime='i1')

    assert list(d.signature.parameters.values()) == [
        Parameter('C', Annotation(arr_type, 'o')),
        Parameter('D', Annotation(arr_type, 'o')),
        Parameter('A_prime', Annotation(arr_type, 'i')),
        Parameter('B', Annotation(arr_type, 'i')),
        Parameter('coeff', Annotation(coeff_dtype))]


def test_alien_parameters():
    """
    Check that one cannot connect transformations using parameter objects from
    other transformation/computation.
    """

    N = 200
    coeff_dtype = numpy.float32
    arr_type = Type(numpy.complex64, (N, N))

    d = Dummy(arr_type, arr_type, coeff_dtype, same_A_B=True)
    d2 = Dummy(arr_type, arr_type, coeff_dtype, same_A_B=True)
    identity = tr_identity(d.parameter.A)
    identity2 = tr_identity(d.parameter.A)

    with pytest.raises(ValueError):
        d.connect(d2.parameter.A, identity, 'o1', A_prime='i1')

    with pytest.raises(ValueError):
        d.connect(d.parameter.A, identity, identity2.o1, A_prime='i1')

    with pytest.raises(ValueError):
        d.parameter.A.connect(identity, identity.o1, A_prime=identity2.i1)


def test_connector_repetition():
    """Check that the connector id cannot be repeated in the connections list."""

    N = 200
    coeff_dtype = numpy.float32
    arr_type = Type(numpy.complex64, (N, N))

    d = Dummy(arr_type, arr_type, coeff_dtype, same_A_B=True)
    identity = tr_identity(d.parameter.A)

    with pytest.raises(ValueError):
        d.parameter.A.connect(identity, identity.o1, A=identity.o1, A_prime=identity.i1)


def test_wrong_connector():
    """Check that the error is thrown if the connector is unknown or is not in the signature."""

    N = 200
    coeff_dtype = numpy.float32
    arr_type = Type(numpy.complex64, (N, N))

    d = Dummy(arr_type, arr_type, coeff_dtype, same_A_B=True)
    identity = tr_identity(d.parameter.A)

    d.parameter.A.connect(identity, identity.o1, A_prime=identity.i1)

    # Connector is missing
    with pytest.raises(ValueError):
        d.connect('AA', identity, identity.o1, A_pp=identity.i1)

    # Node 'A' exists, but it is not a part of the signature
    # (hidden by previously connected transformation).
    with pytest.raises(ValueError):
        d.connect('B', identity, identity.o1, A=identity.i1)


def test_type_mismatch():
    """Check that the error is thrown if the connection is made to an wrong type."""

    N = 200
    coeff_dtype = numpy.float32
    arr_type = Type(numpy.complex64, (N, N))

    d = Dummy(arr_type, arr_type, coeff_dtype, same_A_B=True)
    identity = tr_identity(Type(numpy.complex64, (N, N + 1)))

    with pytest.raises(ValueError):
        d.connect('A', identity, identity.o1, A_prime=identity.i1)


def test_wrong_data_path():
    """
    Check that the error is thrown if the connector is a part of the signature,
    but this particular data path (input or output) is already hidden
    by a previously connected transformation.
    """

    N = 200
    coeff_dtype = numpy.float32
    arr_type = Type(numpy.complex64, (N, N))

    d = DummyAdvanced(arr_type, coeff_dtype)
    identity = tr_identity(d.parameter.C)

    d.parameter.C.connect(identity, identity.o1, C_in=identity.i1)
    d.parameter.D.connect(identity, identity.i1, D_out=identity.o1)
    assert list(d.signature.parameters.values()) == [
        Parameter('C', Annotation(arr_type, 'o')),
        Parameter('C_in', Annotation(arr_type, 'i')),
        Parameter('D_out', Annotation(arr_type, 'o')),
        Parameter('D', Annotation(arr_type, 'i')),
        Parameter('coeff1', Annotation(coeff_dtype)),
        Parameter('coeff2', Annotation(coeff_dtype))]

    # Now input to C is hidden by the previously connected transformation
    with pytest.raises(ValueError):
        d.parameter.C.connect(identity, identity.o1, C_in_prime=identity.i1)

    # Same goes for D
    with pytest.raises(ValueError):
        d.parameter.D.connect(identity, identity.i1, D_out_prime=identity.o1)

    # Also we cannot make one of the transformation outputs an existing output parameter
    with pytest.raises(ValueError):
        d.parameter.C.connect(identity, identity.i1, D_out=identity.o1)

    # Output of C is still available though
    d.parameter.C.connect(identity, identity.i1, C_out=identity.o1)
    assert list(d.signature.parameters.values()) == [
        Parameter('C_out', Annotation(arr_type, 'o')),
        Parameter('C_in', Annotation(arr_type, 'i')),
        Parameter('D_out', Annotation(arr_type, 'o')),
        Parameter('D', Annotation(arr_type, 'i')),
        Parameter('coeff1', Annotation(coeff_dtype)),
        Parameter('coeff2', Annotation(coeff_dtype))]


def test_wrong_transformation_parameters():
    """
    Check that the error is thrown if the list of transformation parameter names
    does not coincide with actual names.
    """

    N = 200
    coeff_dtype = numpy.float32
    arr_type = Type(numpy.complex64, (N, N))

    d = Dummy(arr_type, arr_type, coeff_dtype, same_A_B=True)
    identity = tr_identity(d.parameter.A)

    # ``identity`` does not have ``input`` parameter
    with pytest.raises(ValueError):
        d.parameter.A.connect(identity, identity.o1, A_prime='input')

    # ``identity`` does not have ``i2`` parameter
    with pytest.raises(ValueError):
        d.parameter.A.connect(identity, identity.o1, A_prime=identity.i1, A2='i2')


def test_io_merge(some_thr):
    """
    Check that one can end input and output transformation in the same node
    thus making its role 'io'.
    """

    N = 200
    coeff1 = 3
    coeff2 = 4
    scale_in = 0.5
    scale_out = 0.3
    arr_type = Type(numpy.complex64, (N, N))
    coeff_dtype = numpy.float32

    C = get_test_array_like(arr_type)
    D = get_test_array_like(arr_type)
    C_dev = some_thr.to_device(C)
    D_dev = some_thr.to_device(D)

    d = DummyAdvanced(C, coeff_dtype)
    scale = tr_scale(d.parameter.C, d.parameter.coeff1.dtype)
    d.parameter.C.connect(scale, scale.o1, C_prime=scale.i1, scale_in=scale.s1)
    d.parameter.C.connect(scale, scale.i1, C_prime=scale.o1, scale_out=scale.s1)
    assert list(d.signature.parameters.values()) == [
        Parameter('C_prime', Annotation(arr_type, 'io')),
        Parameter('scale_out', Annotation(coeff_dtype)),
        Parameter('scale_in', Annotation(coeff_dtype)),
        Parameter('D', Annotation(arr_type, 'io')),
        Parameter('coeff1', Annotation(coeff_dtype)),
        Parameter('coeff2', Annotation(coeff_dtype))]

    dc = d.compile(some_thr)
    dc(C_dev, scale_out, scale_in, D_dev, coeff1, coeff2)

    C_ref, D_ref = mock_dummy_advanced(C * scale_in, D, coeff1, coeff2)
    C_ref *= scale_out

    C = C_dev.get()
    D = D_dev.get()

    assert diff_is_negligible(C, C_ref)
    assert diff_is_negligible(D, D_ref)



import numpy

from reikna.cluda import Module
import reikna.helpers as helpers
from reikna.cluda import dtypes


class KeyGenerator:
    """
    Contains a key generator module and accompanying metadata.
    Supports ``__process_modules__`` protocol.

    .. py:attribute:: module

        A module with the key generator function:

    .. c:function:: Key key_from_int(int idx)

        Generates and returns a key, suitable for the bijection which was given to the constructor.
    """

    def __init__(self, module, base_key):
        """__init__()""" # hide the signature from Sphinx
        self.module = module
        self._base_key = base_key

    def __process_modules__(self, process):
        return KeyGenerator(process(self.module), self._base_key)

    @classmethod
    def create(cls, bijection, seed=None, reserve_id_space=True):
        """
        Creates a generator.

        :param bijection: a :py:class:`~reikna.cbrng.bijections.Bijection` object.
        :param seed: an integer, or numpy array of 32-bit unsigned integers.
        :param reserve_id_space: if ``True``, the last 32 bit of the key will be reserved
            for the thread identifier.
            As a result, the total size of the key should be 64 bit or more.
            If ``False``, the thread identifier will be just added to the key,
            which will still result in different keys for different threads,
            with the danger that different seeds produce the same sequences.
        """

        if reserve_id_space:
            if bijection.key_words == 1 and bijection.word_dtype.itemsize == 4:
            # It's too hard to compress both global and thread-dependent part
            # in a single 32-bit word.
            # Let the user handle this himself.
                raise ValueError("Cannor reserve ID space in a 32-bit key")

            if bijection.word_dtype.itemsize == 4:
                key_words32 = bijection.key_words - 1
            else:
                if bijection.key_words > 1:
                    key_words32 = (bijection.key_words - 1) * 2
                else:
                    # Philox-2x64 case, the key is a single 64-bit integer.
                    # We use first 32 bit for the key, and the remaining 32 bit for a thread identifier.
                    key_words32 = 1
        else:
            key_words32 = bijection.key_words * (bijection.word_dtype.itemsize // 4)

        if isinstance(seed, numpy.ndarray):
            # explicit key was provided
            assert seed.size == key_words32 and seed.dtype == numpy.uint32
            key = seed.copy().flatten()
        else:
            # use numpy to generate the key from seed
            np_rng = numpy.random.RandomState(seed)

            # 32-bit Python can only generate random integer up to 2**31-1
            key16 = np_rng.randint(0, 2**16, key_words32 * 2)
            key = numpy.zeros(key_words32, numpy.uint32)
            for i in range(key_words32 * 2):
                key[i // 2] += key16[i] << (16 if i % 2 == 0 else 0)

        full_key = numpy.zeros(1, bijection.key_dtype)[0]
        if bijection.word_dtype.itemsize == 4:
            full_key['v'][:key_words32] = key
        else:
            for i in range(key_words32):
                full_key['v'][i // 2] += key[i] << (32 if i % 2 == 0 else 0)

        module = Module.create("""
            WITHIN_KERNEL ${bijection.module}Key ${prefix}key_from_int(int idx)
            {
                ${bijection.module}Key result;

                %for i in range(bijection.key_words):
                result.v[${i}] = ${key['v'][i]}
                    %if i == bijection.key_words - 1:
                    + idx
                    %endif
                ;
                %endfor

                return result;
            }
            """,
            render_kwds=dict(
                bijection=bijection,
                key=full_key))

        return cls(module, full_key)

    def reference(self, idx):
        """
        Reference function that returns the key given the thread identifier.
        Uses the same algorithm as the module.
        """
        key = self._base_key.copy()
        key['v'][-1] += dtypes.cast(key.dtype.fields['v'][0].base)(idx)
        return key



import reikna.helpers as helpers
from reikna.cluda import Module, Snippet


VALUE_NAME = "_val"
INDEX_NAME = "_idx"


def leaf_name(name):
    return "_leaf_" + name


def index_cnames(shape):
    return [INDEX_NAME + str(i) for i in range(len(shape))]


def index_cnames_seq(param, qualified=False):
    names = index_cnames(param.annotation.type.shape)
    if qualified:
        return ["VSIZE_T " + name for name in names]
    else:
        return names


def flat_index_expr(param):

    type_ = param.annotation.type

    if len(type_.shape) == 0:
        return "0"

    # FIXME: Assuming that all strides are multiples of dtype.itemsize.
    # This can change with custom strides, and we will have
    # to cast device pointer to bytes and back.
    # Need to investigate what happens in this case on some concrete example.
    if not all(stride % type_.dtype.itemsize == 0 for stride in type_.strides):
        raise ValueError(
            "Some of the strides " + str(type_.strides) +
            "are not multiples of the itemsize" + str(type_.dtype.itemsize))

    item_strides = [stride // type_.dtype.itemsize for stride in type_.strides]

    names = index_cnames(param.annotation.type.shape)

    return " + ".join([
        name + " * " + str(stride)
        for name, stride in zip(names, item_strides)])


def param_cname(param, qualified=False):
    # Note that if ``param`` has a struct type,
    # its .annotation.type.ctype attribute can be a module.
    # In that case ``str()`` has to be called explicitly for ``ctype``
    # to get the module prefix.
    name = "_leaf_" + param.name
    if qualified:
        ctype = param.annotation.type.ctype
        if param.annotation.array:
            return "GLOBAL_MEM " + str(ctype) + " *" + name
        else:
            return str(ctype) + " " + name
    else:
        return name


def param_cnames_seq(parameters, qualified=False):
    return [param_cname(p, qualified=qualified) for p in parameters]


_snippet_kernel_declaration = helpers.template_def(
    [],
    "KERNEL void ${kernel_name}(${', '.join(param_cnames_seq(parameters, qualified=True))})")

def kernel_declaration(kernel_name, parameters):
    return Snippet(
        _snippet_kernel_declaration,
        render_kwds=dict(
            param_cnames_seq=param_cnames_seq,
            kernel_name=kernel_name,
            parameters=parameters))


def node_connector(output):
    if output:
        return VALUE_NAME
    else:
        return VALUE_NAME + " ="


_module_transformation = helpers.template_def(
    ['prefix'],
    """
    // ${'output' if output else 'input'} transformation node for "${name}"
    <%
        value_param = [str(connector_ctype) + ' ' + VALUE_NAME] if output else []
        value = [VALUE_NAME] if output else []

        signature = (
            param_cnames_seq(subtree_parameters, qualified=True) +
            q_indices +
            value_param)
    %>
    INLINE WITHIN_KERNEL ${'void' if output else connector_ctype} ${prefix}func(
        ${",\\n".join(signature)})
    {
        %if not output:
        ${connector_ctype} ${VALUE_NAME};
        %endif

        ${tr_snippet(*tr_args)}

        %if not output:
        return ${VALUE_NAME};
        %endif
    }
    <%
    %>
    #define ${prefix}(${', '.join(nq_indices + value)}) ${prefix}func(\\
        ${', '.join(nq_params + nq_indices + value)})
    """)

def module_transformation(output, param, subtree_parameters, tr_snippet, tr_args):
    return Module(
        _module_transformation,
        render_kwds=dict(
            output=output,
            name=param.name,
            param_cnames_seq=param_cnames_seq, subtree_parameters=subtree_parameters,
            q_indices=index_cnames_seq(param, qualified=True),
            VALUE_NAME=VALUE_NAME,
            nq_params=param_cnames_seq(subtree_parameters),
            nq_indices=index_cnames_seq(param),
            connector_ctype=param.annotation.type.ctype,
            tr_snippet=tr_snippet,
            tr_args=tr_args))


_module_leaf_macro = helpers.template_def(
    ['prefix'],
    """
    // leaf ${'output' if output else 'input'} macro for "${name}"
    %if output:
    #define ${prefix}(${', '.join(index_seq + [VALUE_NAME])}) ${lname}[${index_expr}] = (${VALUE_NAME})
    %else:
    #define ${prefix}(${', '.join(index_seq)}) (${lname}[${index_expr}])
    %endif
    """)

def module_leaf_macro(output, param):
    return Module(
        _module_leaf_macro,
        render_kwds=dict(
            output=output,
            name=param.name,
            VALUE_NAME=VALUE_NAME,
            lname=leaf_name(param.name),
            index_seq=index_cnames_seq(param),
            index_expr=flat_index_expr(param)))


_module_same_indices = helpers.template_def(
    ['prefix'],
    """
    // ${'output' if output else 'input'} for a transformation for "${name}"
    %if output:
    #define ${prefix}(${VALUE_NAME}) ${module_idx}(${', '.join(nq_indices + [VALUE_NAME])})
    %else:
    #define ${prefix} ${module_idx}(${', '.join(nq_indices)})
    %endif
    """)

def module_same_indices(output, param, subtree_parameters, module_idx):
    return Module(
        _module_same_indices,
        render_kwds=dict(
            output=output,
            name=param.name,
            VALUE_NAME=VALUE_NAME,
            module_idx=module_idx,
            nq_indices=index_cnames_seq(param),
            nq_params=param_cnames_seq(subtree_parameters)))


_snippet_disassemble_combined = Snippet.create(
    lambda shape, slices, indices, combined_indices: """
    %for combined_index, slice_len in enumerate(slices):
    <%
        index_start = sum(slices[:combined_index])
        index_end = index_start + slice_len
    %>
        %for index in range(index_start, index_end):
        <%
            stride = product(shape[index+1:index_end])
        %>
        VSIZE_T ${indices[index]} = ${combined_indices[combined_index]} / ${stride};
        %if index != index_end - 1:
        ${combined_indices[combined_index]} -= ${indices[index]} * ${stride};
        %endif
        %endfor
    %endfor
    """,
    render_kwds=dict(product=helpers.product))

_module_combined = helpers.template_def(
    ['prefix', 'slices'],
    """
    <%
        value_param = [str(connector_ctype) + ' ' + VALUE_NAME] if output else []
        value = [VALUE_NAME] if output else []

        combined_indices = ['_c_idx' + str(i) for i in range(len(slices))]
        q_combined_indices = ['VSIZE_T ' + ind for ind in combined_indices]
        nq_combined_indices = combined_indices
        signature = (
            param_cnames_str(subtree_parameters, qualified=True) +
            q_combined_indices +
            value_param)
    %>
    INLINE WITHIN_KERNEL ${'void' if output else connector_ctype} ${prefix}func(
        ${', '.join(signature)})
    {
        ${disassemble(shape, slices, indices, combined_indices)}

        %if not output:
        return
        %endif
        ${module_idx}(${', '.join(nq_indices + value)});
    }
    <%
        value = [VALUE_NAME] if output else []
    %>
    #define ${prefix}(${', '.join(nq_combined_indices + value)}) ${prefix}func(\\
        ${', '.join(nq_params + nq_combined_indices + value)})
    """)

def module_combined(output, param, subtree_parameters, module_idx):
    return Module(
        _module_combined,
        render_kwds=dict(
            output=output,
            VALUE_NAME=VALUE_NAME,
            shape=param.annotation.type.shape,
            module_idx=module_idx,
            disassemble=_snippet_disassemble_combined,
            connector_ctype=param.annotation.type.ctype,
            nq_indices=index_cnames_seq(param),
            q_indices=index_cnames_seq(param, qualified=True),
            param_cnames_str=param_cnames_seq, subtree_parameters=subtree_parameters,
            nq_params=param_cnames_seq(subtree_parameters),
            indices=index_cnames(param.annotation.type.shape)))



import itertools
from warnings import catch_warnings, filterwarnings

import pytest

import reikna.cluda as cluda
import reikna.cluda.dtypes as dtypes
import reikna.cluda.functions as functions
from reikna.helpers import product

from helpers import *


def get_func_kernel(thr, func_module, out_dtype, in_dtypes):
    src = """
    <%
        argnames = ["a" + str(i + 1) for i in range(len(in_dtypes))]
        in_ctypes = list(map(dtypes.ctype, in_dtypes))
        out_ctype = dtypes.ctype(out_dtype)
    %>
    KERNEL void test(
        GLOBAL_MEM ${out_ctype} *dest
        %for arg, ctype in zip(argnames, in_ctypes):
        , GLOBAL_MEM ${ctype} *${arg}
        %endfor
        )
    {
        const SIZE_T i = get_global_id(0);
        %for arg, ctype in zip(argnames, in_ctypes):
        ${ctype} ${arg}_load = ${arg}[i];
        %endfor

        dest[i] = ${func}(${", ".join([arg + "_load" for arg in argnames])});
    }
    """

    program = thr.compile(
        src,
        render_kwds=dict(in_dtypes=in_dtypes, out_dtype=out_dtype, func=func_module))

    return program.test


def generate_dtypes(out_code, in_codes):
    test_dtype = lambda idx: dict(i=numpy.int32, f=numpy.float32, c=numpy.complex64)[idx]
    in_dtypes = list(map(test_dtype, in_codes))
    out_dtype = dtypes.result_type(*in_dtypes) if out_code == 'auto' else test_dtype(out_code)

    if not any(map(dtypes.is_double, in_dtypes)):
        # numpy thinks that int32 * float32 == float64,
        # but we still need to run this test on older videocards
        if dtypes.is_complex(out_dtype):
            out_dtype = numpy.complex64
        elif dtypes.is_real(out_dtype):
            out_dtype = numpy.float32

    return out_dtype, in_dtypes


def check_func(thr, func_module, reference_func, out_dtype, in_dtypes):
    N = 256

    test = get_func_kernel(thr, func_module, out_dtype, in_dtypes)

    arrays = [get_test_array(N, dt, no_zeros=True) for dt in in_dtypes]
    arrays_dev = map(thr.to_device, arrays)
    dest_dev = thr.array(N, out_dtype)

    test(dest_dev, *arrays_dev, global_size=N)
    assert diff_is_negligible(thr.from_device(dest_dev), reference_func(*arrays))


@pytest.mark.parametrize(
    ('out_code', 'in_codes'),
    [('f', 'f'), ('c', 'c')])
def test_exp(thr, out_code, in_codes):
    out_dtype, in_dtypes = generate_dtypes(out_code, in_codes)
    check_func(thr, functions.exp(in_dtypes[0]), numpy.exp, out_dtype, in_dtypes)


@pytest.mark.parametrize(
    ('out_code', 'in_codes'),
    [('c', 'f')])
def test_polar_unit(thr, out_code, in_codes):
    out_dtype, in_dtypes = generate_dtypes(out_code, in_codes)
    check_func(
        thr, functions.polar_unit(in_dtypes[0]),
        lambda theta: numpy.exp(1j * theta), out_dtype, in_dtypes)


@pytest.mark.parametrize(
    ('out_code', 'in_codes'),
    [('c', 'ff')])
def test_polar(thr, out_code, in_codes):
    out_dtype, in_dtypes = generate_dtypes(out_code, in_codes)
    check_func(
        thr, functions.polar(in_dtypes[0]),
        lambda rho, theta: rho * numpy.exp(1j * theta), out_dtype, in_dtypes)


@pytest.mark.parametrize(
    ('out_code', 'in_codes'),
    [('f', 'c'), ('f', 'f'), ('i', 'i')])
def test_norm(thr, out_code, in_codes):
    out_dtype, in_dtypes = generate_dtypes(out_code, in_codes)
    check_func(
        thr, functions.norm(in_dtypes[0]),
        lambda x: numpy.abs(x) ** 2, out_dtype, in_dtypes)


@pytest.mark.parametrize(
    ('out_code', 'in_codes'),
    [('c', 'c')])
def test_conj(thr, out_code, in_codes):
    out_dtype, in_dtypes = generate_dtypes(out_code, in_codes)
    check_func(
        thr, functions.conj(in_dtypes[0]),
        numpy.conj, out_dtype, in_dtypes)


@pytest.mark.parametrize(
    ('out_code', 'in_codes'),
    [('c', 'f'), ('f', 'f'), ('c', 'c')])
def test_cast(thr, out_code, in_codes):
    out_dtype, in_dtypes = generate_dtypes(out_code, in_codes)
    check_func(
        thr, functions.cast(out_dtype, in_dtypes[0]),
        dtypes.cast(out_dtype), out_dtype, in_dtypes)


@pytest.mark.parametrize(
    ('out_code', 'in_codes'),
    [('f', 'ff'), ('c', 'cc'), ('c', 'cf'), ('c', 'fc'), ('f', 'if')])
def test_div(thr, out_code, in_codes):
    out_dtype, in_dtypes = generate_dtypes(out_code, in_codes)
    check_func(
        thr, functions.div(*in_dtypes, out_dtype=out_dtype),
        lambda x, y: dtypes.cast(out_dtype)(x / y), out_dtype, in_dtypes)


@pytest.mark.parametrize('in_codes', ["ii", "ff", "cc", "cfi", "ifccfi"])
@pytest.mark.parametrize('out_code', ["auto", "i", "f", "c"])
def test_multiarg_mul(thr, out_code, in_codes):
    """
    Checks multi-argument mul() with a variety of data types.
    """

    out_dtype, in_dtypes = generate_dtypes(out_code, in_codes)

    def reference_mul(*args):
        res = product(args)
        if not dtypes.is_complex(out_dtype) and dtypes.is_complex(res.dtype):
            res = res.real
        return res.astype(out_dtype)

    # Temporarily catching imaginary part truncation warnings
    with catch_warnings():
        filterwarnings("ignore", "", numpy.ComplexWarning)
        mul = functions.mul(*in_dtypes, out_dtype=out_dtype)

    check_func(thr, mul, reference_mul, out_dtype, in_dtypes)


@pytest.mark.parametrize('in_codes', ["ii", "ff", "cc", "cfi", "ifccfi"])
@pytest.mark.parametrize('out_code', ["auto", "i", "f", "c"])
def test_multiarg_add(thr, out_code, in_codes):
    """
    Checks multi-argument add() with a variety of data types.
    """

    out_dtype, in_dtypes = generate_dtypes(out_code, in_codes)

    def reference_add(*args):
        res = sum(args)
        if not dtypes.is_complex(out_dtype) and dtypes.is_complex(res.dtype):
            res = res.real
        return res.astype(out_dtype)

    # Temporarily catching imaginary part truncation warnings
    with catch_warnings():
        filterwarnings("ignore", "", numpy.ComplexWarning)
        mul = functions.add(*in_dtypes, out_dtype=out_dtype)

    check_func(thr, mul, reference_add, out_dtype, in_dtypes)



import time
import itertools

import numpy
import pytest

from helpers import *
from reikna.algorithms import Reduce, Predicate, predicate_sum
from reikna.helpers import template_def
from reikna.cluda import Snippet
import reikna.cluda.dtypes as dtypes


shapes = [
    (2,), (13,), (1535,), (512 * 231,),
    (140, 3), (13, 598), (1536, 789),
    (5, 15, 19), (134, 25, 23), (145, 56, 178)]
shapes_and_axes = [(shape, axis) for shape, axis in itertools.product(shapes, [None, 0, 1, 2])
    if axis is None or axis < len(shape)]
shapes_and_axes_ids = [str(shape) + "," + str(axis) for shape, axis in shapes_and_axes]


@pytest.mark.parametrize(('shape', 'axis'), shapes_and_axes, ids=shapes_and_axes_ids)
def test_normal(thr, shape, axis):

    a = get_test_array(shape, numpy.int64)
    a_dev = thr.to_device(a)

    rd = Reduce(a, predicate_sum(numpy.int64), axes=(axis,) if axis is not None else None)

    b_dev = thr.empty_like(rd.parameter.output)
    b_ref = a.sum(axis)

    rdc = rd.compile(thr)
    rdc(b_dev, a_dev)

    assert diff_is_negligible(b_dev.get(), b_ref)


def test_nondefault_function(thr):

    shape = (100, 100)
    a = get_test_array(shape, numpy.int64)
    a_dev = thr.to_device(a)
    b_ref = a.sum(0)

    predicate = Predicate(
        Snippet.create(lambda v1, v2: "return ${v1} + ${v2};"),
        0)

    rd = Reduce(a_dev, predicate, axes=(0,))

    b_dev = thr.empty_like(rd.parameter.output)

    rdc = rd.compile(thr)
    rdc(b_dev, a_dev)

    assert diff_is_negligible(b_dev.get(), b_ref)


def test_nonsequential_axes(thr):

    shape = (50, 40, 30, 20)
    a = get_test_array(shape, numpy.int64)
    a_dev = thr.to_device(a)
    b_ref = a.sum(0).sum(1) # sum over axes 0 and 2 of the initial array

    rd = Reduce(a_dev, predicate_sum(numpy.int64), axes=(0,2))

    b_dev = thr.empty_like(rd.parameter.output)

    rdc = rd.compile(thr)
    rdc(b_dev, a_dev)

    assert diff_is_negligible(b_dev.get(), b_ref)


def test_structure_type(thr):

    shape = (100, 100)
    dtype = dtypes.align(numpy.dtype([
        ('i1', numpy.uint32),
        ('nested', numpy.dtype([
            ('v', numpy.uint64),
            ])),
        ('i2', numpy.uint32)
        ]))

    a = get_test_array(shape, dtype)
    a_dev = thr.to_device(a)

    # Have to construct the resulting array manually,
    # since numpy cannot reduce arrays with struct dtypes.
    b_ref = numpy.empty(100, dtype)
    b_ref['i1'] = a['i1'].sum(0)
    b_ref['nested']['v'] = a['nested']['v'].sum(0)
    b_ref['i2'] = a['i2'].sum(0)

    predicate = Predicate(
        Snippet.create(lambda v1, v2: """
            ${ctype} result = ${v1};
            result.i1 += ${v2}.i1;
            result.nested.v += ${v2}.nested.v;
            result.i2 += ${v2}.i2;
            return result;
            """,
            render_kwds=dict(
                ctype=dtypes.ctype_module(dtype))),
        numpy.zeros(1, dtype)[0])

    rd = Reduce(a_dev, predicate, axes=(0,))

    b_dev = thr.empty_like(rd.parameter.output)

    rdc = rd.compile(thr)
    rdc(b_dev, a_dev)

    assert diff_is_negligible(b_dev.get(), b_ref)


@pytest.mark.perf
@pytest.mark.returns('GB/s')
def test_summation(thr):

    perf_size = 2 ** 22
    dtype = dtypes.normalize_type(numpy.int64)

    a = get_test_array(perf_size, dtype)
    a_dev = thr.to_device(a)

    rd = Reduce(a, predicate_sum(dtype))

    b_dev = thr.empty_like(rd.parameter.output)
    b_ref = numpy.array([a.sum()], dtype)

    rdc = rd.compile(thr)

    attempts = 10
    times = []
    for i in range(attempts):
        t1 = time.time()
        rdc(b_dev, a_dev)
        thr.synchronize()
        times.append(time.time() - t1)

    assert diff_is_negligible(b_dev.get(), b_ref)

    return min(times), perf_size * dtype.itemsize




import numpy

from reikna.cluda import Snippet
import reikna.helpers as helpers
from reikna.cluda import dtypes
from reikna.cluda import OutOfResourcesError
from reikna.core import Computation, Parameter, Annotation, Type
from reikna.algorithms import Transpose

TEMPLATE = helpers.template_for(__file__)


class Predicate:
    """
    A predicate used in :py:class:`~reikna.algorithms.Reduce`.

    :param operation: a :py:class:`~reikna.cluda.Snippet` object with two parameters
        which will take the names of two arguments to join.
    :param empty: a numpy scalar with the empty value of the argument
        (the one which, being joined by another argument, does not change it).
    """

    def __init__(self, operation, empty):
        self.operation = operation
        self.empty = empty

    def __process_modules__(self, process):
        return Predicate(process(self.operation), self.empty)


def predicate_sum(dtype):
    """
    Returns a :py:class:`~reikna.algorithms.Predicate` object which sums its arguments.
    """
    return Predicate(
        Snippet.create(lambda v1, v2: "return ${v1} + ${v2};"),
        numpy.zeros(1, dtype)[0])


class Reduce(Computation):
    """
    Bases: :py:class:`~reikna.core.Computation`

    Reduces the array over given axis using given binary operation.

    :param arr_t: an array-like defining the initial array.
    :param predicate: a :py:class:`~reikna.algorithms.Predicate` object.
    :param axes: a list of non-repeating axes to reduce over.
        If ``None``, the whole array will be reduced
        (in which case the shape of the output array is ``(1,)``).

    .. py:method:: compiled_signature(output:o, input:i)

        :param input: an array with the attributes of ``arr_t``.
        :param output: an array with the attributes of ``arr_t``,
            with its shape missing axes from ``axes``.
    """

    def __init__(self, arr_t, predicate, axes=None):

        dims = len(arr_t.shape)

        if axes is None:
            axes = tuple(range(dims))
        else:
            axes = tuple(sorted(helpers.wrap_in_tuple(axes)))

        if len(set(axes)) != len(axes):
            raise ValueError("Cannot reduce twice over the same axis")

        if min(axes) < 0 or max(axes) >= dims:
            raise ValueError("Axes numbers are out of bounds")

        if hasattr(predicate.empty, 'dtype'):
            if arr_t.dtype != predicate.empty.dtype:
                raise ValueError("The predicate and the array must use the same data type")
            empty = predicate.empty
        else:
            empty = dtypes.cast(arr_t.dtype)(predicate.empty)

        remaining_axes = tuple(a for a in range(dims) if a not in axes)
        output_shape = tuple(arr_t.shape[a] for a in remaining_axes)

        if axes == tuple(range(dims - len(axes), dims)):
            self._transpose_axes = None
        else:
            self._transpose_axes = remaining_axes + axes

        self._operation = predicate.operation
        self._empty = empty

        Computation.__init__(self, [
            Parameter('output', Annotation(Type(arr_t.dtype, shape=output_shape), 'o')),
            Parameter('input', Annotation(arr_t, 'i'))])

    def _build_plan_for_wg_size(self, plan_factory, warp_size, max_wg_size, output, input_):

        plan = plan_factory()

        # Using algorithm cascading: sequential reduction, and then the parallel one.
        # According to Brent's theorem, the optimal sequential size is O(log(n)).
        # Setting it to the nearest power of 2 to simplify integer operations.
        max_seq_size = helpers.bounding_power_of_2(helpers.log2(max_wg_size))
        max_reduce_power = max_wg_size * max_seq_size

        if self._transpose_axes is None:
            # normal reduction
            cur_input = input_
        else:
            transpose = Transpose(input_, axes=self._transpose_axes)
            tr_output = plan.temp_array_like(transpose.parameter.output)
            plan.computation_call(transpose, tr_output, input_)

            cur_input = tr_output

        axis_start = len(output.shape)
        axis_end = len(input_.shape) - 1

        input_slices = (axis_start, axis_end - axis_start + 1)

        part_size = helpers.product(cur_input.shape[axis_start:])
        final_size = helpers.product(cur_input.shape[:axis_start])

        while part_size > 1:

            if part_size > max_reduce_power:
                seq_size = max_seq_size
                block_size = max_wg_size
                blocks_per_part = helpers.min_blocks(part_size, block_size * seq_size)
                cur_output = plan.temp_array(
                    (final_size, blocks_per_part), input_.dtype)
                output_slices = (1, 1)
            else:
                if part_size > max_wg_size:
                    seq_size = helpers.min_blocks(part_size, max_wg_size)
                    block_size = max_wg_size
                else:
                    seq_size = 1
                    block_size = helpers.bounding_power_of_2(part_size)
                blocks_per_part = 1
                cur_output = output
                output_slices = (len(cur_output.shape), 0)

            if part_size % (block_size * seq_size) != 0:
                last_block_size = part_size % (block_size * seq_size)
            else:
                last_block_size = block_size * seq_size

            render_kwds = dict(
                seq_size=seq_size,
                blocks_per_part=blocks_per_part,
                last_block_size=last_block_size,
                log2=helpers.log2, block_size=block_size,
                warp_size=warp_size,
                empty=self._empty,
                operation=self._operation,
                input_slices=input_slices,
                output_slices=output_slices)

            plan.kernel_call(
                TEMPLATE.get_def('reduce'),
                [cur_output, cur_input],
                global_size=(final_size, blocks_per_part * block_size),
                local_size=(1, block_size),
                render_kwds=render_kwds)

            part_size = blocks_per_part
            cur_input = cur_output
            input_slices = output_slices

        return plan

    def _build_plan(self, plan_factory, device_params, output, input_):

        max_wg_size = device_params.max_work_group_size

        while max_wg_size >= 1:

            try:
                plan = self._build_plan_for_wg_size(
                    plan_factory, device_params.warp_size, max_wg_size, output, input_)
            except OutOfResourcesError:
                max_wg_size /= 2
                continue

            return plan

        raise ValueError("Could not find suitable call parameters for one of the local kernels")



import numpy
import pytest

from helpers import *
from reikna.algorithms import PureParallel
from reikna.core import Parameter, Annotation, Type, Computation
import reikna.cluda.dtypes as dtypes
from reikna.transformations import mul_param


class NestedPureParallel(Computation):

    def __init__(self, size, dtype):

        Computation.__init__(self, [
            Parameter('output', Annotation(Type(dtype, shape=size), 'o')),
            Parameter('input', Annotation(Type(dtype, shape=size), 'i'))])

        self._p = PureParallel([
                Parameter('output', Annotation(Type(dtype, shape=size), 'o')),
                Parameter('i1', Annotation(Type(dtype, shape=size), 'i')),
                Parameter('i2', Annotation(Type(dtype, shape=size), 'i'))],
            """
            ${i1.ctype} t1 = ${i1.load_idx}(${idxs[0]});
            ${i2.ctype} t2 = ${i2.load_idx}(${idxs[0]});
            ${output.store_idx}(${idxs[0]}, t1 + t2);
            """)

    def _build_plan(self, plan_factory, device_params, output, input_):
        plan = plan_factory()
        plan.computation_call(self._p, output, input_, input_)
        return plan


def test_nested(thr):

    N = 1000
    dtype = numpy.float32

    p = NestedPureParallel(N, dtype)

    a = get_test_array_like(p.parameter.input)
    a_dev = thr.to_device(a)
    res_dev = thr.empty_like(p.parameter.output)

    pc = p.compile(thr)
    pc(res_dev, a_dev)

    res_ref = a + a

    assert diff_is_negligible(res_dev.get(), res_ref)


def test_guiding_input(thr):

    N = 1000
    dtype = numpy.float32

    p = PureParallel(
        [
            Parameter('output', Annotation(Type(dtype, shape=(2, N)), 'o')),
            Parameter('input', Annotation(Type(dtype, shape=N), 'i'))],
        """
        float t = ${input.load_idx}(${idxs[0]});
        ${output.store_idx}(0, ${idxs[0]}, t);
        ${output.store_idx}(1, ${idxs[0]}, t * 2);
        """,
        guiding_array='input')

    a = get_test_array_like(p.parameter.input)
    a_dev = thr.to_device(a)
    res_dev = thr.empty_like(p.parameter.output)

    pc = p.compile(thr)
    pc(res_dev, a_dev)

    res_ref = numpy.vstack([a, a * 2])

    assert diff_is_negligible(res_dev.get(), res_ref)


def test_guiding_output(thr):

    N = 1000
    dtype = numpy.float32

    p = PureParallel(
        [
            Parameter('output', Annotation(Type(dtype, shape=N), 'o')),
            Parameter('input', Annotation(Type(dtype, shape=(2, N)), 'i'))],
        """
        float t1 = ${input.load_idx}(0, ${idxs[0]});
        float t2 = ${input.load_idx}(1, ${idxs[0]});
        ${output.store_idx}(${idxs[0]}, t1 + t2);
        """,
        guiding_array='output')

    a = get_test_array_like(p.parameter.input)
    a_dev = thr.to_device(a)
    res_dev = thr.empty_like(p.parameter.output)

    pc = p.compile(thr)
    pc(res_dev, a_dev)

    res_ref = a[0] + a[1]

    assert diff_is_negligible(res_dev.get(), res_ref)


def test_guiding_shape(thr):

    N = 1000
    dtype = numpy.float32

    p = PureParallel(
        [
            Parameter('output', Annotation(Type(dtype, shape=(2, N)), 'o')),
            Parameter('input', Annotation(Type(dtype, shape=(2, N)), 'i'))],
        """
        float t1 = ${input.load_idx}(0, ${idxs[0]});
        float t2 = ${input.load_idx}(1, ${idxs[0]});
        ${output.store_idx}(0, ${idxs[0]}, t1 + t2);
        ${output.store_idx}(1, ${idxs[0]}, t1 - t2);
        """,
        guiding_array=(N,))

    a = get_test_array_like(p.parameter.input)
    a_dev = thr.to_device(a)
    res_dev = thr.empty_like(p.parameter.output)

    pc = p.compile(thr)
    pc(res_dev, a_dev)

    res_ref = numpy.vstack([a[0] + a[1], a[0] - a[1]])

    assert diff_is_negligible(res_dev.get(), res_ref)


def test_zero_length_shape(thr):

    dtype = numpy.float32

    p = PureParallel(
        [
            Parameter('output', Annotation(Type(dtype, shape=tuple()), 'o')),
            Parameter('input', Annotation(Type(dtype, shape=tuple()), 'i'))],
        """
        float t = ${input.load_idx}();
        ${output.store_idx}(t * 2);
        """,
        guiding_array=tuple())

    a = get_test_array_like(p.parameter.input)
    a_dev = thr.to_device(a)
    res_dev = thr.empty_like(p.parameter.output)

    pc = p.compile(thr)
    pc(res_dev, a_dev)

    res_ref = (a * 2).astype(dtype)

    assert diff_is_negligible(res_dev.get(), res_ref)


def test_trf_with_guiding_input(thr):
    """
    Test the creation of ``PureParallel`` out of a transformation,
    with an input parameter as a guiding array.
    """

    N = 1000
    coeff = 3
    dtype = numpy.float32

    arr_t = Type(dtype, shape=N)
    trf = mul_param(arr_t, dtype)
    p = PureParallel.from_trf(trf, trf.input)

    # The new PureParallel has to preserve the parameter list of the original transformation.
    assert list(p.signature.parameters.values()) == list(trf.signature.parameters.values())

    a = get_test_array_like(p.parameter.input)
    a_dev = thr.to_device(a)
    res_dev = thr.empty_like(p.parameter.output)

    pc = p.compile(thr)
    pc(res_dev, a_dev, coeff)

    assert diff_is_negligible(res_dev.get(), a * 3)


def test_trf_with_guiding_output(thr):
    """
    Test the creation of ``PureParallel`` out of a transformation,
    with an output parameter as a guiding array.
    """

    N = 1000
    coeff = 3
    dtype = numpy.float32

    arr_t = Type(dtype, shape=N)
    trf = mul_param(arr_t, dtype)
    p = PureParallel.from_trf(trf, trf.output)

    # The new PureParallel has to preserve the parameter list of the original transformation.
    assert list(p.signature.parameters.values()) == list(trf.signature.parameters.values())

    a = get_test_array_like(p.parameter.input)
    a_dev = thr.to_device(a)
    res_dev = thr.empty_like(p.parameter.output)

    pc = p.compile(thr)
    pc(res_dev, a_dev, coeff)

    assert diff_is_negligible(res_dev.get(), a * 3)



from reikna.core import Computation, Parameter, Annotation, Type
from reikna.transformations import norm_const
from reikna.algorithms import Reduce, predicate_sum


class EntrywiseNorm(Computation):
    r"""
    Bases: :py:class:`~reikna.core.Computation`

    Calculates the entrywise matrix norm (same as ``numpy.linalg.norm``)
    of an arbitrary order :math:`r`:

    .. math::

        ||A||_r = \left( \sum_{i,j,\ldots} |A_{i,j,\ldots}|^r \right)^{1 / r}

    :param arr_t: an array-like defining the initial array.
    :param order: the order :math:`r` (any real number).
    :param axes: a list of non-repeating axes to sum over.
        If ``None``, the norm of the whole array will be calculated.

    .. py:method:: compiled_signature(output:o, input:i)

        :param input: an array with the attributes of ``arr_t``.
        :param output: an array with the attributes of ``arr_t``,
            with its shape missing axes from ``axes``.
    """

    def __init__(self, arr_t, order=2, axes=None):
        tr_elems = norm_const(arr_t, order)
        out_dtype = tr_elems.output.dtype
        res_t = Type(out_dtype)
        tr_sum = norm_const(res_t, 1. / order)

        rd = Reduce(Type(out_dtype, arr_t.shape), predicate_sum(out_dtype), axes=axes)
        rd.parameter.input.connect(tr_elems, tr_elems.output, input_prime=tr_elems.input)
        rd.parameter.output.connect(tr_sum, tr_sum.input, output_prime=tr_sum.output)

        self._rd = rd

        Computation.__init__(self, [
            Parameter('output', Annotation(res_t, 'o')),
            Parameter('input', Annotation(arr_t, 'i'))])

    def _build_plan(self, plan_factory, device_params, output, input_):
        plan = plan_factory()
        plan.computation_call(self._rd, output, input_)
        return plan



"""
This module contains a number of pre-created transformations.
"""

import reikna.cluda.dtypes as dtypes
import reikna.cluda.functions as functions
from reikna.core import Transformation, Parameter, Annotation, Type


def copy(arr_t, out_arr_t=None):
    """
    Returns an identity transformation (1 output, 1 input): ``output = input``.
    Output array type ``out_arr_t`` may have different strides,
    but must have the same shape and data type.
    """
    if out_arr_t is None:
        out_arr_t = arr_t
    else:
        if out_arr_t.shape != arr_t.shape or out_arr_t.dtype != arr_t.dtype:
            raise ValueError("Input and output arrays must have the same shape and data type")

    return Transformation(
        [Parameter('output', Annotation(out_arr_t, 'o')),
        Parameter('input', Annotation(arr_t, 'i'))],
        "${output.store_same}(${input.load_same});")


def add_param(arr_t, param_dtype):
    """
    Returns an addition transformation with a dynamic parameter (1 output, 1 input, 1 scalar):
    ``output = input + param``.
    """
    return Transformation(
        [Parameter('output', Annotation(arr_t, 'o')),
        Parameter('input', Annotation(arr_t, 'i')),
        Parameter('param', Annotation(param_dtype))],
        "${output.store_same}(${add}(${input.load_same}, ${param}));",
        render_kwds=dict(add=functions.add(arr_t.dtype, param_dtype, out_dtype=arr_t.dtype)))


def add_const(arr_t, param):
    """
    Returns an addition transformation with a fixed parameter (1 output, 1 input):
    ``output = input + param``.
    """
    param_dtype = dtypes.detect_type(param)
    return Transformation(
        [Parameter('output', Annotation(arr_t, 'o')),
        Parameter('input', Annotation(arr_t, 'i'))],
        "${output.store_same}(${add}(${input.load_same}, ${param}));",
        render_kwds=dict(
            add=functions.add(arr_t.dtype, param_dtype, out_dtype=arr_t.dtype),
            param=dtypes.c_constant(param, dtype=param_dtype)))


def mul_param(arr_t, param_dtype):
    """
    Returns a scaling transformation with a dynamic parameter (1 output, 1 input, 1 scalar):
    ``output = input * param``.
    """
    return Transformation(
        [Parameter('output', Annotation(arr_t, 'o')),
        Parameter('input', Annotation(arr_t, 'i')),
        Parameter('param', Annotation(param_dtype))],
        "${output.store_same}(${mul}(${input.load_same}, ${param}));",
        render_kwds=dict(mul=functions.mul(arr_t.dtype, param_dtype, out_dtype=arr_t.dtype)))


def mul_const(arr_t, param):
    """
    Returns a scaling transformation with a fixed parameter (1 output, 1 input):
    ``output = input * param``.
    """
    param_dtype = dtypes.detect_type(param)
    return Transformation(
        [Parameter('output', Annotation(arr_t, 'o')),
        Parameter('input', Annotation(arr_t, 'i'))],
        "${output.store_same}(${mul}(${input.load_same}, ${param}));",
        render_kwds=dict(
            mul=functions.mul(arr_t.dtype, param_dtype, out_dtype=arr_t.dtype),
            param=dtypes.c_constant(param, dtype=param_dtype)))


def split_complex(input_arr_t):
    """
    Returns a transformation that splits complex input into two real outputs
    (2 outputs, 1 input): ``real = Re(input), imag = Im(input)``.
    """
    output_t = Type(dtypes.real_for(input_arr_t.dtype), shape=input_arr_t.shape)
    return Transformation(
        [Parameter('real', Annotation(output_t, 'o')),
        Parameter('imag', Annotation(output_t, 'o')),
        Parameter('input', Annotation(input_arr_t, 'i'))],
        """
            ${real.store_same}(${input.load_same}.x);
            ${imag.store_same}(${input.load_same}.y);
        """)


def combine_complex(output_arr_t):
    """
    Returns a transformation that joins two real inputs into complex output
    (1 output, 2 inputs): ``output = real + 1j * imag``.
    """
    input_t = Type(dtypes.real_for(output_arr_t.dtype), shape=output_arr_t.shape)
    return Transformation(
        [Parameter('output', Annotation(output_arr_t, 'o')),
        Parameter('real', Annotation(input_t, 'i')),
        Parameter('imag', Annotation(input_t, 'i'))],
        """
        ${output.store_same}(
            COMPLEX_CTR(${output.ctype})(
                ${real.load_same},
                ${imag.load_same}));
        """)


def norm_const(arr_t, order):
    """
    Returns a transformation that calculates the ``order``-norm
    (1 output, 1 input): ``output = abs(input) ** order``.
    """
    if dtypes.is_complex(arr_t.dtype):
        out_dtype = dtypes.real_for(arr_t.dtype)
    else:
        out_dtype = arr_t.dtype

    return Transformation(
        [
            Parameter('output', Annotation(Type(out_dtype, arr_t.shape), 'o')),
            Parameter('input', Annotation(arr_t, 'i'))],
        """
        ${input.ctype} val = ${input.load_same};
        ${output.ctype} norm = ${norm}(val);
        %if order != 2:
        norm = pow(norm, ${dtypes.c_constant(order / 2, output.dtype)});
        %endif
        ${output.store_same}(norm);
        """,
        render_kwds=dict(
            norm=functions.norm(arr_t.dtype),
            order=order))


def norm_param(arr_t):
    """
    Returns a transformation that calculates the ``order``-norm
    (1 output, 1 input, 1 param): ``output = abs(input) ** order``.
    """
    if dtypes.is_complex(arr_t.dtype):
        out_dtype = dtypes.real_for(arr_t.dtype)
    else:
        out_dtype = arr_t.dtype

    return Transformation(
        [
            Parameter('output', Annotation(Type(out_dtype, arr_t.shape), 'o')),
            Parameter('input', Annotation(arr_t, 'i')),
            Parameter('order', Annotation(Type(out_dtype)))],
        """
        ${input.ctype} val = ${input.load_same};
        ${output.ctype} norm = ${norm}(val);
        norm = pow(norm, ${order} / 2);
        ${output.store_same}(norm);
        """,
        render_kwds=dict(
            norm=functions.norm(arr_t.dtype)))


def ignore(arr_t):
    """
    Returns a transformation that ignores the output it is attached to.
    """
    return Transformation(
        [Parameter('input', Annotation(arr_t, 'i'))],
        """
        // Ignoring intentionally
        """)


def broadcast_const(arr_t, val):
    """
    Returns a transformation that broadcasts the given constant to the array output
    (1 output): ``output = val``.
    """
    val = dtypes.cast(arr_t.dtype)(val)
    if len(val.shape) != 0:
        raise ValueError("The constant must be a scalar")
    return Transformation(
        [
            Parameter('output', Annotation(arr_t, 'o'))],
        """
        const ${output.ctype} val = ${dtypes.c_constant(val)};
        ${output.store_same}(val);
        """,
        render_kwds=dict(val=val))


def broadcast_param(arr_t):
    """
    Returns a transformation that broadcasts the free parameter to the array output
    (1 output, 1 param): ``output = param``.
    """
    return Transformation(
        [
            Parameter('output', Annotation(arr_t, 'o')),
            Parameter('param', Annotation(Type(arr_t.dtype)))],
        """
        ${output.store_same}(${param});
        """)



"""
Reference implementation of the counter-based RNGs Threefry and Philox from
Salmon et al., P. Int. C. High. Perform. 16 (2011), doi:10.1145/2063384.2063405.
Based on the source code of Random123 library (http://www.thesalmons.org/john/random123/).

This implementation favors simplicity over speed and therefore
is not for use in production.
It only guarantees to produce the same results as the original Random123.
"""
from __future__ import print_function
import sys

import numpy

from reikna.helpers import ignore_integer_overflow


major, _, _, _, _ = sys.version_info
if major < 3:
    long_int = long
else:
    long_int = int


# Rotation constants:
THREEFRY_ROTATION = {
    # These are the R_256 constants from the Threefish reference sources
    # with names changed to R_64x4...
    (64, 4): numpy.array([[14, 52, 23, 5, 25, 46, 58, 32], [16, 57, 40, 37, 33, 12, 22, 32]]).T,

    # Output from skein_rot_search: (srs64_B64-X1000)
    # Random seed = 1. BlockSize = 128 bits. sampleCnt =  1024. rounds =  8, minHW_or=57
    # Start: Tue Mar  1 10:07:48 2011
    # rMin = 0.136. #0325[*15] [CRC=455A682F. hw_OR=64. cnt=16384. blkSize= 128].format
    (64, 2): numpy.array([[16, 42, 12, 31, 16, 32, 24, 21]]).T,
    # 4 rounds: minHW =  4  [  4  4  4  4 ]
    # 5 rounds: minHW =  8  [  8  8  8  8 ]
    # 6 rounds: minHW = 16  [ 16 16 16 16 ]
    # 7 rounds: minHW = 32  [ 32 32 32 32 ]
    # 8 rounds: minHW = 64  [ 64 64 64 64 ]
    # 9 rounds: minHW = 64  [ 64 64 64 64 ]
    # 10 rounds: minHW = 64  [ 64 64 64 64 ]
    # 11 rounds: minHW = 64  [ 64 64 64 64 ]

    # Output from skein_rot_search: (srs-B128-X5000.out)
    # Random seed = 1. BlockSize = 64 bits. sampleCnt =  1024. rounds =  8, minHW_or=28
    # Start: Mon Aug 24 22:41:36 2009
    # ...
    # rMin = 0.472. #0A4B[*33] [CRC=DD1ECE0F. hw_OR=31. cnt=16384. blkSize= 128].format
    (32, 4): numpy.array([[10, 11, 13, 23, 6, 17, 25, 18], [26, 21, 27, 5, 20, 11, 10, 20]]).T,
    # 4 rounds: minHW =  3  [  3  3  3  3 ]
    # 5 rounds: minHW =  7  [  7  7  7  7 ]
    # 6 rounds: minHW = 12  [ 13 12 13 12 ]
    # 7 rounds: minHW = 22  [ 22 23 22 23 ]
    # 8 rounds: minHW = 31  [ 31 31 31 31 ]
    # 9 rounds: minHW = 32  [ 32 32 32 32 ]
    # 10 rounds: minHW = 32  [ 32 32 32 32 ]
    # 11 rounds: minHW = 32  [ 32 32 32 32 ]

    # Output from skein_rot_search (srs32x2-X5000.out)
    # Random seed = 1. BlockSize = 64 bits. sampleCnt =  1024. rounds =  8, minHW_or=28
    # Start: Tue Jul 12 11:11:33 2011
    # rMin = 0.334. #0206[*07] [CRC=1D9765C0. hw_OR=32. cnt=16384. blkSize=  64].format
    (32, 2): numpy.array([[13, 15, 26, 6, 17, 29, 16, 24]]).T
    # 4 rounds: minHW =  4  [  4  4  4  4 ]
    # 5 rounds: minHW =  6  [  6  8  6  8 ]
    # 6 rounds: minHW =  9  [  9 12  9 12 ]
    # 7 rounds: minHW = 16  [ 16 24 16 24 ]
    # 8 rounds: minHW = 32  [ 32 32 32 32 ]
    # 9 rounds: minHW = 32  [ 32 32 32 32 ]
    # 10 rounds: minHW = 32  [ 32 32 32 32 ]
    # 11 rounds: minHW = 32  [ 32 32 32 32 ]

}


THREEFRY_KS_PARITY = {
    64: 0x1BD11BDAA9FC1A22,
    32: 0x1BD11BDA
}


def threefry_rotate(W, R, x):
    # Cast to uint is required by numpy coercion rules.
    # "% W" is technically redundant since R < W always.
    s1 = numpy.uint32(R % W)
    s2 = numpy.uint32((W - R) % W)
    return (x << s1) | (x >> s2)


def threefry(W, N, ctr, key, Nrounds=None):
    """
    W: word length (32, 64)
    N: number of generated items, 2 or 4
    Nrounds: number of rounds (up to 72)
    ctr: counter, array(N, numpy.uint${W})
    key: key, array(N, numpy.uint${W})
    returns: array(N, numpy.uint${W})

    Note: for W=64, N=4 and Nrounds=72 it is the same as Threefish algorithm from Skein,
    only without the 128-bit "tweak" which is applied to the key.
    With this tweak it is possible to upgrade this algorithm from PRNG to QRNG.
    """
    assert W in (32, 64)
    assert N in (2, 4)

    if Nrounds is None:
        Nrounds = 20

    dtype = numpy.uint32 if W == 32 else numpy.uint64

    if N == 2:
        assert Nrounds <= 32
        ks = numpy.empty(3, dtype)
    else:
        assert Nrounds <= 72
        ks = numpy.empty(5, dtype)

    X = numpy.empty(N, dtype)
    assert ctr.size == key.size == N

    ks[N] = THREEFRY_KS_PARITY[W]
    for i in range(N):
        ks[i] = key[i]
        X[i] = ctr[i]
        ks[N] ^= key[i]

    # Insert initial key before round 0
    for i in range(N):
        X[i] += ks[i]

    R = THREEFRY_ROTATION[(W, N)]

    with ignore_integer_overflow():
        for rnd in range(Nrounds):

            # FIXME: In the current version of Random123 (1.06),
            # there is a bug in R_idx calculation for N == 2, where
            #    R_idx = rnd % 8 if rnd < 20 else (rnd - 4) % 8
            # instead of what goes below.
            # When this bug is fixed, Random123 and this implementation
            # will start to produce identical results again,
            # and this comment can be removed.
            R_idx = rnd % 8

            if N == 2:
                X[0] += X[1]
                X[1] = threefry_rotate(W, R[R_idx, 0], X[1])
                X[1] ^= X[0]
            else:
                idx1 = 1 if rnd % 2 == 0 else 3
                idx2 = 3 if rnd % 2 == 0 else 1

                X[0] += X[idx1]
                X[idx1] = threefry_rotate(W, R[R_idx, 0], X[idx1])
                X[idx1] ^= X[0]

                X[2] += X[idx2]
                X[idx2] = threefry_rotate(W, R[R_idx, 1], X[idx2])
                X[idx2] ^= X[2]

            if rnd % 4 == 3:
                for i in range(N):
                    X[i] += ks[(rnd // 4 + i + 1) % (N + 1)]

                X[N-1] += dtype(rnd // 4 + 1)

    return X


PHILOX_W = {
    64: [
        numpy.uint64(0x9E3779B97F4A7C15), # golden ratio
        numpy.uint64(0xBB67AE8584CAA73B) # sqrt(3)-1
    ],
    32: [
        numpy.uint32(0x9E3779B9), # golden ratio
        numpy.uint32(0xBB67AE85) # sqrt(3)-1
    ]
}

PHILOX_M = {
    (64,2): [numpy.uint64(0xD2B74407B1CE6E93)],
    (64,4): [numpy.uint64(0xD2E7470EE14C6C93), numpy.uint64(0xCA5A826395121157)],
    (32,2): [numpy.uint32(0xD256D193)],
    (32,4): [numpy.uint32(0xD2511F53), numpy.uint32(0xCD9E8D57)]
}


def philox_mulhilo(W, x, y):
    res = long_int(x) * long_int(y)
    cast = numpy.cast[x.dtype]
    return cast(res // (2 ** W)), cast(res % (2 ** W))


def philox_round(W, N, rnd, ctr, key):
    ctr = ctr.copy()
    rnd = numpy.cast[ctr.dtype](rnd)

    if N == 2:
        key0 = key[0] + PHILOX_W[W][0] * rnd
        hi, lo = philox_mulhilo(W, PHILOX_M[(W,N)][0], ctr[0])
        ctr[0] = hi ^ key0 ^ ctr[1]
        ctr[1] = lo
    else:
        key0 = key[0] + PHILOX_W[W][0] * rnd
        key1 = key[1] + PHILOX_W[W][1] * rnd
        hi0, lo0 = philox_mulhilo(W, PHILOX_M[(W,N)][0], ctr[0])
        hi1, lo1 = philox_mulhilo(W, PHILOX_M[(W,N)][1], ctr[2])
        ctr[0] = hi1 ^ ctr[1] ^ key0
        ctr[1] = lo1
        ctr[2] = hi0 ^ ctr[3] ^ key1
        ctr[3] = lo0

    return ctr


def philox(W, N, ctr, key, Nrounds=None):
    """
    W: word length (32, 64)
    N: number of generated items, 2 or 4
    Nrounds: number of rounds (up to 16)
    inn: counter, array(N, numpy.uint${W}) --- counter
    k: key, array(N/2, numpy.uint${W}) --- key
    returns: array(N, numpy.uint${W})
    """
    assert W in (32, 64)
    assert N in (2, 4)

    if Nrounds is None:
        Nrounds = 10

    assert Nrounds <= 16

    with ignore_integer_overflow():
        for rnd in range(Nrounds):
            ctr = philox_round(W, N, rnd, ctr, key)

    return ctr


# Some simple testcases to compare with Random123 results.

def myhex(x):
    s = hex(x)[2:-1]
    return "0" * (2 * x.nbytes - len(s)) + s


def test(name):
    kk = numpy.array([123, 456, 789, 101112])

    for W in (32, 64):
        for N in (2, 4):

            dtype = numpy.uint32 if W == 32 else numpy.uint64
            if name == 'threefry':
                key = kk[:N].astype(dtype)
                rounds = 72 if N == 4 else 32
                func = threefry
            elif name == 'philox':
                key = kk[:N//2].astype(dtype)
                rounds = 10
                func = philox

            ctr = numpy.zeros(N).astype(dtype)

            print(name, ": W =", W, "N =", N)
            for i in range(5):
                ctr[0] = i
                res = func(W, N, ctr, key, Nrounds=rounds)
                print("ctr " + str(i) + ": " + " ".join([myhex(i) for i in res]))


if __name__ == '__main__':
    test('threefry')
    test('philox')




import itertools
import time

import numpy
import pytest

from helpers import *

from reikna.helpers import product
from reikna.fft import FFT
import reikna.cluda.dtypes as dtypes
from reikna.transformations import mul_param


def pytest_generate_tests(metafunc):

    perf_log_shapes = [
        (4,), (10,), (13,), # 1D
        (4, 4), (7, 7), (10, 10), # 2D
        (4, 4, 4), (5, 5, 7), (7, 7, 7)] # 3D
    perf_mem_limit = 4 * 2**20

    if 'local_shape_and_axes' in metafunc.funcargnames:
        def idgen(val):
            batch, size = val[0]
            return str(batch) + 'x' + str(size)

        # These values are supposed to check all code paths in
        # fft.mako::insertGlobalLoadsAndTranspose.
        # Some of them will probably become global FFTs on lower-end GPUs, but that's fine.
        #
        # We need to try different FFT sizes in order to catch the path where
        # 1. ``threads_per_xform >= mem_coalesce_width``
        #    (first code path, corresponds to relatively large FFT sizes)
        # 2. ``xforms_per_workgroup`` still > 1
        #    (which will allow us to have ``xforms_remainder != 0``)

        mem_limit = 2 ** 20
        size_powers = (3, 7, 8, 9, 10, 12)

        vals = []
        for p in size_powers:
            size = 2 ** p
            batch = mem_limit // size

            # "size / 2 - 1" will lead to full FFT of size ``size``
            # in the current version of Bluestein's algorithm
            pad_test_size = size // 2 - 1

            for s in (size, pad_test_size):
                # testing batch = 1, non-multiple batch and power of 2 batch
                for b in (1, batch - 1, batch):
                    vals.append(((b, s), (1,)))

        metafunc.parametrize('local_shape_and_axes', vals, ids=list(map(idgen, vals)))

    elif 'global_shape_and_axes' in metafunc.funcargnames:
        def idgen(val):
            outer_batch, size, inner_batch = val[0]
            return str(outer_batch) + 'x' + str(size) + 'x' + str(inner_batch)

        # These values are supposed to check all code paths in
        # fft.mako::fft_global.
        mem_limit = 2 ** 22
        size_powers = (3, 7, 9, 10)

        vals = []
        for p in size_powers:
            size = 2 ** p
            batch = mem_limit // size // 64

            # "size / 2 - 1" will lead to full FFT of size ``size``
            # in the current version of Bluestein's algorithm
            pad_test_size = size // 2 - 1

            for s in (size, pad_test_size):
                # possible inner batches: small power of 2, small non-power-of-2,
                # big power of 2 (more than coalesce_width), big non-power-of-2
                for ib in (2, 3, 63, 64):
                    for ob in (1, batch - 1, batch):
                        vals.append(((ob, s, ib), (1,)))

        # big size (supposed to be > local_kernel_limit * MAX_RADIX)
        big_size = 2 ** 15
        batch = mem_limit // big_size // 8
        vals.append(((1, big_size, 1), (1,)))
        vals.append(((1, big_size, 2), (1,)))
        vals.append(((1, big_size, 3), (1,)))
        vals.append(((batch, big_size, 3), (1,)))
        vals.append(((batch - 1, big_size, 3), (1,)))

        metafunc.parametrize('global_shape_and_axes', vals, ids=list(map(idgen, vals)))

    elif 'sequence_shape_and_axes' in metafunc.funcargnames:

        def idgen(non2problem_shape_and_axes):
            shape, axes = non2problem_shape_and_axes
            return str(shape) + 'over' + str(axes)

        vals = [
            ((3, 255, 7), (2, 0)),
            ((17, 200, 131), (1, 0)),
            ((7, 1000, 11), (1, 2)),
            ((15, 900, 57), (2, 0, 1))]

        metafunc.parametrize('sequence_shape_and_axes', vals, ids=list(map(idgen, vals)))

    elif 'perf_shape_and_axes' in metafunc.funcargnames:

        vals = []
        ids = []
        for log_shape in perf_log_shapes:
            shape = tuple(2 ** x for x in log_shape)
            batch = perf_mem_limit // (2 ** sum(log_shape))
            vals.append(((batch,) + shape, tuple(range(1, len(shape) + 1))))
            ids.append(str(batch) + "x" + str(shape))

        metafunc.parametrize('perf_shape_and_axes', vals, ids=ids)

    elif 'non2problem_perf_shape_and_axes' in metafunc.funcargnames:

        vals = []
        ids = []
        for log_shape in perf_log_shapes:
            for modifier in (1, -1):
                shape = tuple(2 ** (x - 1) + modifier for x in log_shape)
                batch = perf_mem_limit // (2 ** sum(log_shape))
                vals.append(((batch,) + shape, tuple(range(1, len(shape) + 1))))
                ids.append(str(batch) + "x" + str(shape))

        metafunc.parametrize('non2problem_perf_shape_and_axes', vals, ids=ids)


def check_errors(thr, shape_and_axes):

    dtype = numpy.complex64

    shape, axes = shape_and_axes

    data = get_test_array(shape, dtype)

    fft = FFT(data, axes=axes)
    fftc = fft.compile(thr)

    # forward transform
    # Testing inplace transformation, because if this works,
    # then the out of place one will surely work too.
    data_dev = thr.to_device(data)
    fftc(data_dev, data_dev)
    fwd_ref = numpy.fft.fftn(data, axes=axes).astype(dtype)
    assert diff_is_negligible(data_dev.get(), fwd_ref)

    # inverse transform
    data_dev = thr.to_device(data)
    fftc(data_dev, data_dev, inverse=True)
    inv_ref = numpy.fft.ifftn(data, axes=axes).astype(dtype)
    assert diff_is_negligible(data_dev.get(), inv_ref)


def test_trivial(some_thr):
    """
    Checks that even if the FFT is trivial (problem size == 1),
    the transformations are still attached and executed.
    """
    dtype = numpy.complex64
    shape = (128, 1, 1, 128)
    axes = (1, 2)
    param = 4

    data = get_test_array(shape, dtype)
    data_dev = some_thr.to_device(data)
    res_dev = some_thr.empty_like(data_dev)

    fft = FFT(data_dev, axes=axes)
    scale = mul_param(data_dev, numpy.int32)
    fft.parameter.input.connect(scale, scale.output, input_prime=scale.input, param=scale.param)

    fftc = fft.compile(some_thr)
    fftc(res_dev, data_dev, param)
    assert diff_is_negligible(res_dev.get(), data * param)


def test_local(thr, local_shape_and_axes):
    check_errors(thr, local_shape_and_axes)

def test_global(thr, global_shape_and_axes):
    check_errors(thr, global_shape_and_axes)

def test_sequence(thr, sequence_shape_and_axes):
    check_errors(thr, sequence_shape_and_axes)


def check_performance(thr_and_double, shape_and_axes, fast_math):
    thr, double = thr_and_double

    shape, axes = shape_and_axes
    dtype = numpy.complex128 if double else numpy.complex64

    data = get_test_array(shape, dtype)
    data_dev = thr.to_device(data)
    res_dev = thr.empty_like(data_dev)

    fft = FFT(data_dev, axes=axes)
    fftc = fft.compile(thr, fast_math=fast_math)

    attempts = 10
    t1 = time.time()
    for i in range(attempts):
        fftc(res_dev, data_dev)
    thr.synchronize()
    t2 = time.time()
    dev_time = (t2 - t1) / attempts

    fwd_ref = numpy.fft.fftn(data, axes=axes).astype(dtype)
    assert diff_is_negligible(res_dev.get(), fwd_ref)

    return dev_time, product(shape) * sum([numpy.log2(shape[a]) for a in axes]) * 5


@pytest.mark.perf
@pytest.mark.returns('GFLOPS')
def test_power_of_2_performance(thr_and_double, perf_shape_and_axes, fast_math):
    return check_performance(thr_and_double, perf_shape_and_axes, fast_math)


@pytest.mark.perf
@pytest.mark.returns('GFLOPS')
def test_non_power_of_2_performance(thr_and_double, non2problem_perf_shape_and_axes, fast_math):
    return check_performance(thr_and_double, non2problem_perf_shape_and_axes, fast_math)



import collections
import weakref

from reikna.helpers.sortedcollection import SortedCollection


def extract_dependencies(dependencies):
    """
    Recursively extracts allocation identifiers from an iterable or Array class.
    """
    results = set()

    if isinstance(dependencies, collections.Iterable):
        for dep in dependencies:
            results.update(extract_dependencies(dep))
    elif hasattr(dependencies, '__tempalloc__'):
        # hook for exposing temporary allocations in arbitrary classes
        results.update(extract_dependencies(dependencies.__tempalloc__))
    elif hasattr(dependencies, '__tempalloc_id__'):
        return set([dependencies.__tempalloc_id__])

    return results


class TemporaryManager:
    """
    Base class for a manager of temporary allocations.

    :param thr: an instance of :py:class:`~reikna.cluda.api.Thread`.
    :param pack_on_alloc: whether to repack allocations when a new allocation is requested.
    :param pack_on_free: whether to repack allocations when an allocation is freed.
    """

    def __init__(self, thr, pack_on_alloc=False, pack_on_free=False):
        self._thr = thr
        self._id_counter = 0
        self._arrays = {}
        self._pack_on_alloc = pack_on_alloc
        self._pack_on_free = pack_on_free

    def array(self, shape, dtype, strides=None, dependencies=None):
        """
        Returns a temporary array.

        :param shape: shape of the array.
        :param dtype: data type of the array.
        :param strides: tuple of bytes to step in each dimension when traversing an array.
        :param dependencies: can be a :py:class:`~reikna.cluda.api.Array` instance
            (the ones containing persistent allocations will be ignored),
            an iterable with valid values,
            or an object with the attribute `__tempalloc__` which is a valid value
            (the last two will be processed recursively).
        """

        # Used to hook memory allocation in the array constructor
        # and save the requested raw memory size.
        class DummyAllocator:
            def __init__(self):
                self.size = None
            def __call__(self, size):
                self.size = size

        new_id = self._id_counter
        self._id_counter += 1

        allocator = DummyAllocator()
        array = self._thr.array(shape, dtype, strides=strides, allocator=allocator)
        array.__tempalloc_id__ = new_id

        dependencies = extract_dependencies(dependencies)
        self._allocate(new_id, allocator.size, dependencies, self._pack_on_alloc)
        self._arrays[new_id] = weakref.ref(array, lambda _: self.free(new_id))

        if self._pack_on_alloc:
            self.update_all()
        else:
            self.update_buffer(new_id)

        return array

    def update_buffer(self, id_):
        array = self._arrays[id_]()
        buf = self._get_buffer(id_)
        if hasattr(array, 'base_data'):
            array.base_data = buf
        else:
            array.gpudata = buf

    def update_all(self):
        for id_ in self._arrays:
            self.update_buffer(id_)

    def free(self, id_):
        array = self._arrays[id_]()
        if array is not None:
            raise Exception("Attempting to free the buffer of an existing temporary array")

        del self._arrays[id_]
        self._free(id_, self._pack_on_free)
        if self._pack_on_free:
            self.update_all()

    def pack(self):
        """
        Packs the real allocations possibly reducing total memory usage.
        This process can be slow.
        """
        self._pack()
        self.update_all()


class TrivialManager(TemporaryManager):
    """
    Trivial manager --- allocates a separate buffer for each allocation request.
    """

    def __init__(self, *args, **kwds):
        TemporaryManager.__init__(self, *args, **kwds)
        self._allocations = {}

    def _allocate(self, new_id, size, _dependencies, _pack):
        buf = self._thr.allocate(size)
        self._allocations[new_id] = buf

    def _get_buffer(self, id_):
        return self._allocations[id_]

    def _free(self, id_, _pack):
        del self._allocations[id_]

    def _pack(self):
        pass


class ZeroOffsetManager(TemporaryManager):
    """
    Tries to assign several allocation requests to a single real allocation,
    if dependencies allow that.
    All virtual allocations start from the beginning of real allocations.
    """

    VirtualAllocation = collections.namedtuple('VirtualAllocation', ['size', 'dependencies'])
    RealAllocation = collections.namedtuple('RealAllocation', ['buffer', 'virtual_ids'])
    RealSize = collections.namedtuple('RealSize', ['size', 'real_id'])
    VirtualMapping = collections.namedtuple('VirtualMapping', ['real_id', 'sub_region'])

    def __init__(self, *args, **kwds):
        TemporaryManager.__init__(self, *args, **kwds)

        self._virtual_allocations = {} # id -> VirtualAllocation
        self._real_sizes = SortedCollection(key=lambda x: x.size) # RealSize objects, sorted by size
        self._virtual_to_real = {} # id -> VirtualMapping
        self._real_allocations = {} # real_id -> RealAllocation
        self._real_id_counter = 0

    def _allocate(self, new_id, size, dependencies, pack):

        # Dependencies should be bidirectional.
        # So if some new allocation says it depends on earlier ones,
        # we need to update their dependency lists.
        dep_set = set(dependencies)
        for dep in dependencies:
            if dep in self._virtual_allocations:
                self._virtual_allocations[dep].dependencies.add(new_id)
            else:
                dep_set.remove(dep)

        # Save virtual allocation parameters
        self._virtual_allocations[new_id] = self.VirtualAllocation(size, dep_set)

        if pack:
            # If pack is requested, we can just do full re-pack right away.
            self._pack()
        else:
            # If not, find a real allocation using the greedy algorithm.
            self._fast_add(new_id, size, dep_set)

    def _fast_add(self, new_id, size, dep_set):
        """
        Greedy algorithm to find a real allocation for a given virtual allocation.
        """

        # Find the smallest real allocation which can hold the requested virtual allocation.
        try:
            idx_start = self._real_sizes.argfind_ge(size)
        except ValueError:
            idx_start = len(self._real_sizes)

        # Check all real allocations with suitable sizes, starting from the smallest one.
        # Use the first real allocation which does not contain ``new_id``'s dependencies.
        for idx in range(idx_start, len(self._real_sizes)):
            real_id = self._real_sizes[idx].real_id
            buf = self._real_allocations[real_id].buffer
            virtual_ids = self._real_allocations[real_id].virtual_ids
            if virtual_ids.isdisjoint(dep_set):
                virtual_ids.add(new_id)
                break
        else:
            # If no suitable real allocation is found, create a new one.
            buf = self._thr.allocate(size)
            real_id = self._real_id_counter
            self._real_id_counter += 1

            self._real_allocations[real_id] = self.RealAllocation(buf, set([new_id]))
            self._real_sizes.insert(self.RealSize(size, real_id))

        # Here it would be more appropriate to use buffer.get_sub_region(0, size),
        # but OpenCL does not allow several overlapping subregions to be used in a single kernel
        # for both read and write, which ruins the whole idea.
        # So we are passing full buffers and hope that overlapping Array class takes care of sizes.
        self._virtual_to_real[new_id] = self.VirtualMapping(
            real_id, self._real_allocations[real_id].buffer)

    def _get_buffer(self, id_):
        return self._virtual_to_real[id_].sub_region

    def _free(self, id_, pack=False):
        # Remove the allocation from the dependency lists of its dependencies
        dep_set = self._virtual_allocations[id_].dependencies
        for dep in dep_set:
            self._virtual_allocations[dep].dependencies.remove(id_)

        vtr = self._virtual_to_real[id_]

        # Clear virtual allocation data
        del self._virtual_allocations[id_]
        del self._virtual_to_real[id_]

        if pack:
            self._pack()
        else:
            # Fast and non-optimal free.
            # Remove the virtual allocation from the real allocation,
            # and delete the real allocation if its no longer used by other virtual allocations.
            ra = self._real_allocations[vtr.real_id]
            ra.virtual_ids.remove(id_)
            if len(ra.virtual_ids) == 0:
                del self._real_allocations[vtr.real_id]
                self._real_sizes.remove(self.RealSize(ra.buffer.size, vtr.real_id))

    def _pack(self):
        """
        Full memory re-pack.
        In theory, should find the optimal (with the minimal real allocation size) distribution
        of virtual allocations.
        """

        # Need to synchronize, because we are going to change allocation addresses,
        # and we do not want to free the memory some kernel is reading from.
        self._thr.synchronize()

        # Clear all real allocation data.
        self._real_sizes.clear()
        self._real_allocations = {}
        self._real_id_counter = 0

        va = self._virtual_allocations

        # Sort all virtual allocations by size
        virtual_sizes = sorted(
            [(va[id_].size, id_) for id_ in va],
            key=lambda x: x[0])

        # Application of greedy algorithm for virtual allocations starting from the largest one
        # should give the optimal distribution.
        for size, id_ in reversed(virtual_sizes):
            self._fast_add(id_, size, self._virtual_allocations[id_].dependencies)

    def _statistics(self):

        stats = dict(
            virtual_size_total=0,
            virtual_num=0,
            real_size_total=0,
            real_num=0,
            virtual_sizes=[],
            real_sizes=[])

        for va in self._virtual_allocations.values():
            stats['virtual_size_total'] += va.size
            stats['virtual_num'] += 1
            stats['virtual_sizes'].append(va.size)

        for ra in self._real_allocations.values():
            stats['real_size_total'] += ra.buffer.size
            stats['real_num'] += 1
            stats['real_sizes'].append(ra.buffer.size)

        stats['virtual_sizes'] = sorted(stats['virtual_sizes'])
        stats['real_sizes'] = sorted(stats['real_sizes'])

        return stats



import numpy
import pytest

from helpers import *
from reikna.linalg import EntrywiseNorm


@pytest.mark.parametrize('dtype', [numpy.float32, numpy.complex64], ids=['float32', 'complex64'])
@pytest.mark.parametrize('order', [0.5, 1, 2])
def test_entrywise_norm(thr, dtype, order):

    a = get_test_array(1000, dtype)
    a_dev = thr.to_device(a)

    norm = EntrywiseNorm(a_dev, order=order)

    b_dev = thr.empty_like(norm.parameter.output)
    b_ref = numpy.linalg.norm(a, ord=order).astype(norm.parameter.output.dtype)

    normc = norm.compile(thr)
    normc(b_dev, a_dev)

    assert diff_is_negligible(b_dev.get(), b_ref)



"""
.. py:class:: Buffer

    Low-level untyped memory allocation.
    Actual class depends on the API: ``pycuda.driver.DeviceAllocation`` for ``CUDA``
    and ``pyopencl.Buffer`` for ``OpenCL``.

    .. py:attribute:: size

.. py:class:: Array

    Actual array class is different depending on the API: ``pycuda.gpuarray.GPUArray``
    for ``CUDA`` and ``pyopencl.array.Array`` for ``OpenCL``.
    This is an interface they both provide.

    .. py:attribute:: shape

    .. py:attribute:: dtype

    .. py:method:: get()

        Returns ``numpy.ndarray`` with the contents of the array.
        Synchronizes the parent :py:class:`~reikna.cluda.api.Thread`.

.. py:class:: DeviceParameters(device)

    An assembly of device parameters necessary for optimizations.

    .. py:attribute:: max_work_group_size

        Maximum block size for kernels.

    .. py:attribute:: max_work_item_sizes

        List with maximum local_size for each dimension.

    .. py:attribute:: max_num_groups

        List with maximum number of workgroups for each dimension.

    .. py:attribute:: warp_size

        Warp size (nVidia), or wavefront size (AMD), or SIMD width is supposed to be
        the number of threads that are executed simultaneously on the same computation unit
        (so you can assume that they are perfectly synchronized).

    .. py:attribute:: local_mem_banks

        Number of local (shared in CUDA) memory banks is a number of successive 32-bit words
        you can access without getting bank conflicts.

    .. py:attribute:: local_mem_size

        Size of the local (shared in CUDA) memory per workgroup, in bytes.

    .. py:attribute:: min_mem_coalesce_width

        Dictionary ``{word_size:elements}``, where ``elements`` is the number of elements
        with size ``word_size`` in global memory that allow coalesced access.

    .. py:method:: supports_dtype(self, dtype)

        Checks if given ``numpy`` dtype can be used in kernels compiled using this thread.

.. py:class:: Platform

    A vendor-specific implementation of the GPGPU API.

    .. py:attribute:: name

        Platform name.

    .. py:attribute:: vendor

        Vendor name.

    .. py:attribute:: version

        Platform version.

    .. py:method:: get_devices()

        Returns a list of device objects available in the platform.
"""

from __future__ import print_function
from logging import error
import weakref
import sys

from reikna.cluda import find_devices
from reikna.helpers import product
from reikna.cluda.kernel import render_prelude, render_template_source
from reikna.cluda.vsize import VirtualSizes
from reikna.cluda.tempalloc import ZeroOffsetManager

_input = input if sys.version_info[0] >= 3 else raw_input


def get_id():
    """
    Returns the identifier of this API.
    """
    raise NotImplementedError()


def get_platforms():
    """
    Returns a list of available :py:class:`Platform` objects.
    In case of ``OpenCL`` returned objects are actually instances of ``pyopencl.Platform``.
    """
    raise NotImplementedError()


class Thread:
    """
    Wraps an existing context in the CLUDA thread object.

    :param cqd: a ``Context``, ``Device`` or ``Stream``/``CommandQueue`` object to base on.
        If a context is passed, a new stream/queue will be created internally.
    :param async: whether to execute all operations with this thread asynchronously
        (you would generally want to set it to ``False`` only for profiling purposes).

    .. note::
        If you are using ``CUDA`` API, you must keep in mind the stateful nature of CUDA calls.
        Briefly, this means that there is the context stack, and the current context on top of it.
        When the :py:meth:`create` is called, the ``PyCUDA`` context gets pushed to the stack
        and made current.
        When the thread object goes out of scope (and the thread object owns it),
        the context is popped, and it is the user's responsibility to make sure
        the popped context is the correct one.
        In simple single-context programs this only means that one should avoid reference cycles
        involving the thread object.

    .. warning::

        Do not pass one ``Stream``/``CommandQueue`` object to several ``Thread`` objects.

    .. py:attribute:: api

        Module object representing the CLUDA API corresponding to this ``Thread``.

    .. py:attribute:: device_params

        Instance of :py:class:`DeviceParameters` class for this thread's device.

    .. py:attribute:: temp_alloc

        Instance of :py:class:`~reikna.cluda.tempalloc.TemporaryManager`
        which handles allocations of temporary arrays (see :py:meth:`temp_array`).
    """

    @classmethod
    def create(cls, interactive=False, device_filters=None, **thread_kwds):
        """
        Creates a new ``Thread`` object with its own context and queue inside.
        Intended for cases when you want to base your whole program on CLUDA.

        :param interactive: ask a user to choose a platform and a device from the ones found.
            If there is only one platform/device available, they will be chosen automatically.
        :param device_filters: keywords to filter devices
            (see the keywords for :py:func:`~reikna.cluda.find_devices`).
        :param thread_kwds: keywords to pass to :py:class:`Thread` constructor.
        :param kwds: same as in :py:class:`Thread`.
        """

        if device_filters is None:
            device_filters = {}
        devices = find_devices(cls.api, **device_filters)

        platforms = cls.api.get_platforms()

        if interactive:
            pnums = sorted(devices.keys())
            if len(pnums) == 1:
                selected_pnum = pnums[0]
                print("Platform:", platforms[0].name)
            else:
                print("Platforms:")
                default_pnum = pnums[0]
                for pnum in pnums:
                    print("[{pnum}]: {pname}".format(pnum=pnum, pname=platforms[pnum].name))
                print(
                    "Choose the platform [{default_pnum}]:".format(default_pnum=default_pnum),
                    end='')
                selected_pnum = _input()
                if selected_pnum == '':
                    selected_pnum = default_pnum
                else:
                    selected_pnum = int(selected_pnum)

            platform = platforms[selected_pnum]
            dnums = sorted(devices[selected_pnum])
            if len(dnums) == 1:
                selected_dnum = dnums[0]
                print("Device:", platform.get_devices()[0].name)
            else:
                print("Devices:")
                default_dnum = dnums[0]
                for dnum in dnums:
                    print("[{dnum}]: {dname}".format(
                        dnum=dnum, dname=platform.get_devices()[dnum].name))
                print(
                    "Choose the device [{default_dnum}]:".format(default_dnum=default_dnum),
                    end='')
                selected_dnum = _input()
                if selected_dnum == '':
                    selected_dnum = default_dnum
                else:
                    selected_dnum = int(selected_dnum)

        else:
            selected_pnum = sorted(devices.keys())[0]
            selected_dnum = devices[selected_pnum][0]

        if thread_kwds is None:
            thread_kwds = {}
        return cls(platforms[selected_pnum].get_devices()[selected_dnum], **thread_kwds)

    def __init__(self, cqd, async=True, temp_alloc=None):

        self._released = False
        self._async = async

        # Make the fields initialized even in case _prcess_cqd() raises an exception.
        self._context = None
        self._queue = None
        self._device = None
        self._owns_context = False
        self._context, self._queue, self._device, self._owns_context = self._process_cqd(cqd)

        self.device_params = self.api.DeviceParameters(self._device)

        temp_alloc_params = dict(
            cls=ZeroOffsetManager, pack_on_alloc=False, pack_on_free=False)
        if temp_alloc is not None:
            temp_alloc_params.update(temp_alloc)

        self.temp_alloc = temp_alloc_params['cls'](weakref.proxy(self),
            pack_on_alloc=temp_alloc_params['pack_on_alloc'],
            pack_on_free=temp_alloc_params['pack_on_free'])

    def allocate(self, size):
        """
        Creates an untyped memory allocation object of type :py:class:`Buffer` with size ``size``.
        """
        raise NotImplementedError()

    def array(self, shape, dtype, strides=None, allocator=None):
        """
        Creates an :py:class:`Array` on GPU with given ``shape``, ``dtype`` and ``strides``.
        Optionally, an ``allocator`` is a callable returning any object castable to ``int``
        representing the physical address on the device (for instance, :py:class:`Buffer`).
        """
        raise NotImplementedError()

    def temp_array(self, shape, dtype, strides=None, dependencies=None):
        """
        Creates an :py:class:`Array` on GPU with given ``shape``, ``dtype`` and ``strides``.
        In order to reduce the memory footprint of the program, the temporary array manager
        will allow these arrays to overlap.
        Two arrays will not overlap, if one of them was specified in ``dependencies``
        for the other one.
        For a list of values ``dependencies`` takes, see the reference entry for
        :py:class:`~reikna.cluda.tempalloc.TemporaryManager`.
        """
        return self.temp_alloc.array(shape, dtype, strides=strides, dependencies=dependencies)

    def empty_like(self, arr):
        """
        Allocates an array on GPU with the same attributes as ``arr``.
        """
        if hasattr(arr, 'allocator'):
            allocator = arr.allocator
        else:
            allocator = None
        return self.array(arr.shape, arr.dtype, strides=arr.strides, allocator=allocator)

    def to_device(self, arr, dest=None):
        """
        Copies an array to the device memory.
        If ``dest`` is specified, it is used as the destination, and the method returns ``None``.
        Otherwise the destination array is created internally and returned from the method.
        """
        if dest is None:
            arr_device = self.empty_like(arr)
        else:
            arr_device = dest

        self._copy_array(arr_device, arr)
        self._synchronize()

        if dest is None:
            return arr_device

    def from_device(self, arr, dest=None, async=False):
        """
        Transfers the contents of ``arr`` to a ``numpy.ndarray`` object.
        The effect of ``dest`` parameter is the same as in :py:meth:`to_device`.
        If ``async`` is ``True``, the transfer is asynchronous
        (the thread-wide asynchronisity setting does not apply here).

        Alternatively, one can use :py:meth:`Array.get`.
        """
        raise NotImplementedError()

    def copy_array(self, arr, dest=None, src_offset=0, dest_offset=0, size=None):
        """
        Copies array on device.

        :param dest: the effect is the same as in :py:meth:`to_device`.
        :param src_offset: offset (in items of ``arr.dtype``) in the source array.
        :param dest_offset: offset (in items of ``arr.dtype``) in the destination array.
        :param size: how many elements of ``arr.dtype`` to copy.
        """

        if dest is None:
            arr_device = self.empty_like(arr)
        else:
            arr_device = dest

        itemsize = arr.dtype.itemsize
        nbytes = arr.nbytes if size is None else itemsize * size
        src_offset *= itemsize
        dest_offset *= itemsize

        self._copy_array_buffer(arr_device, arr,
            nbytes, src_offset=src_offset, dest_offset=dest_offset)
        self._synchronize()

        if dest is None:
            return arr_device

    def synchronize(self):
        """
        Forcefully synchronize this thread with the main program.
        """
        raise NotImplementedError()

    def _synchronize(self):
        if not self._async:
            self.synchronize()

    def _create_program(self, src, fast_math=False):
        try:
            program = self._compile(src, fast_math=fast_math)
        except:
            listing = "\n".join([str(i+1) + ":" + l for i, l in enumerate(src.split('\n'))])
            error("Failed to compile:\n" + listing)
            raise
        return program

    def compile(self, template_src, render_args=None, render_kwds=None, fast_math=False):
        """
        Creates a module object from the given template.

        :param template_src: Mako template source to render
        :param render_kwds: an iterable with positional arguments to pass to the template.
        :param render_kwds: a dictionary with keyword parameters to pass to the template.
        :param fast_math: whether to enable fast mathematical operations during compilation.
        :returns: a :py:class:`Program` object.
        """
        src = render_template_source(
            template_src, render_args=render_args, render_kwds=render_kwds)
        return Program(self, src, fast_math=fast_math)

    def compile_static(self, template_src, name, global_size,
            local_size=None, render_args=None, render_kwds=None, fast_math=False):
        """
        Creates a kernel object with fixed call sizes,
        which allows to overcome some backend limitations.
        Global and local sizes can have any length, providing that
        ``len(global_size) >= len(local_size)``, and the total number of work items and work groups
        is less than the corresponding total number available for the device.
        In order to get IDs and sizes in such kernels, virtual size functions have to be used
        (see :c:macro:`VIRTUAL_SKIP_THREADS` and others for details).

        :param template_src: Mako template or a template source to render
        :param name: name of the kernel function
        :param global_size: global size to be used, in **row-major** order.
        :param local_size: local size to be used, in **row-major** order.
            If ``None``, some suitable one will be picked.
        :param local_mem: (**CUDA API only**) amount of dynamically allocated local memory
            to be used (in bytes).
        :param render_args: a list of parameters to be passed as positional arguments
            to the template.
        :param render_kwds: a dictionary with additional parameters
            to be used while rendering the template.
        :param fast_math: whether to enable fast mathematical operations during compilation.
        :returns: a :py:class:`StaticKernel` object.
        """
        return StaticKernel(self, template_src, name, global_size,
            local_size=local_size, render_args=render_args, render_kwds=render_kwds,
            fast_math=fast_math)

    def _release_specific(self):
        """
        Overridden by a specific ``Thread`` if it needs to do something before finalizing.
        """
        pass

    def release(self):
        """
        Forcefully free critical resources (rendering the object unusable).
        In most cases you can rely on the garbage collector taking care of things.
        Calling this method explicitly may be necessary in case of CUDA API
        when you want to make sure the context got popped.
        """
        # Note that we rely on reference counting here, which is not available
        # in some Python implementations.
        # But since PyCUDA and PyOpenCL only work in CPython anyway,
        # and I cannot see the way to make Thread release both GC-only-compatible
        # and auto-releasable when not in use (crucial for CUDA because of its
        # stupid stateful contexts), I'll leave at is it is for now.

        if not self._released:
            self._release_specific()
            del self._device
            del self._queue
            del self._context
            self._released = True


class Program:
    """
    An object with compiled GPU code.

    .. py:attribute:: source

        Contains module source code.

    .. py:attribute:: kernel_name

        Contains :py:class:`Kernel` object for the kernel ``kernel_name``.
    """

    def __init__(self, thr, src, static=False, fast_math=False):
        """__init__()""" # hide the signature from Sphinx

        self._thr = thr
        self._static = static

        prelude = render_prelude(self._thr, fast_math=fast_math)

        # Casting source code to ASCII explicitly
        # New versions of Mako produce Unicode output by default,
        # and it makes the compiler unhappy
        self.source = str(prelude + src)
        self._program = thr._create_program(self.source, fast_math=fast_math)

    def __getattr__(self, name):
        return self._thr.api.Kernel(self._thr, self._program, name, static=self._static)


class Kernel:
    """
    An object containing GPU kernel.

    .. py:attribute:: max_work_group_size

        Maximum size of the work group for the kernel.
    """

    def __init__(self, thr, program, name, static=False):
        """__init__()""" # hide the signature from Sphinx

        self._thr = thr
        self._program = program
        self._kernel = self._get_kernel(program, name)
        self._static = static
        self._fill_attributes()

    def prepare(self, global_size, local_size=None, local_mem=0):
        """
        Prepare the kernel for execution with given parameters.

        :param global_size: an integer or a tuple of integers,
            specifying total number of work items to run.
        :param local_size: an integer or a tuple of integers,
            specifying the size of a single work group.
            Should have the same number of dimensions as ``global_size``.
            If ``None`` is passed, some ``local_size`` will be picked internally.
        :param local_mem: (**CUDA API only**) amount of dynamic local memory (in bytes)
        """
        raise NotImplementedError()

    def prepared_call(self, *args):
        """
        Execute the kernel.
        :py:class:`Array` objects are allowed as arguments.
        """
        self._prepared_call(*args)
        self._thr._synchronize()

    def __call__(self, *args, **kwds):
        """
        A shortcut for successive call to :py:meth:`prepare` and :py:meth:`prepared_call`.
        """
        if not self._static:
            if 'global_size' in kwds:
                prep_args = (kwds.pop('global_size'),)
            else:
                raise TypeError("global_size keyword argument must be set")
            self.prepare(*prep_args, **kwds)

        self.prepared_call(*args)


class StaticKernel:
    """
    An object containing a GPU kernel with fixed call sizes.

    .. py:attribute:: source

        Contains the source code of the program.
    """

    def __init__(self, thr, template_src, name, global_size, local_size=None,
            render_args=None, render_kwds=None, fast_math=False):
        """__init__()""" # hide the signature from Sphinx

        self._thr = thr

        if render_args is None:
            render_args = []
        if render_kwds is None:
            render_kwds = {}

        main_src = render_template_source(
            template_src, render_args=render_args, render_kwds=render_kwds)

        # Since virtual size function require some registers, they affect the maximum local size.
        # Start from the device's max work group size as the first approximation
        # and recompile kernels with smaller local sizes until convergence.
        max_local_size = thr.device_params.max_work_group_size

        while True:

            # Try to find kernel launch parameters for the requested local size.
            # May raise OutOfResourcesError if it's not possible,
            # just let it pass to the caller.
            vs = VirtualSizes(
                thr.device_params, global_size,
                virtual_local_size=local_size,
                max_local_size=max_local_size)

            # Try to compile the kernel with the corresponding virtual size functions
            program = Program(
                self._thr, vs.vsize_functions + main_src,
                static=True, fast_math=fast_math)
            kernel = getattr(program, name)

            if kernel.max_work_group_size >= product(vs.real_local_size):
                # Kernel will execute with this local size, use it
                break

            # By the contract of VirtualSizes,
            # product(vs.real_local_size) <= max_local_size
            # Also, since we're still in this loop,
            # kernel.max_work_group_size < product(vs.real_local_size).
            # Therefore the new max_local_size value is guaranteed
            # to be smaller than the previous one.
            max_local_size = kernel.max_work_group_size

        self._program = program
        self._kernel = kernel
        self.virtual_local_size = vs.virtual_local_size
        self.virtual_global_size = vs.virtual_global_size
        self.local_size = vs.real_local_size
        self.global_size = vs.real_global_size

        self._kernel.prepare(self.global_size, local_size=self.local_size)

    def __call__(self, *args):
        """
        Execute the kernel.
        :py:class:`Array` objects are allowed as arguments.
        """
        self._kernel(*args)



import numpy

import reikna.helpers as helpers
from reikna.core import Computation, Parameter, Annotation, Type

TEMPLATE = helpers.template_for(__file__)


def transpose_shape(shape, axes):
    return tuple(shape[i] for i in axes)

def transpose_axes(axes, b_start, c_start):
    return axes[:b_start] + axes[c_start:] + axes[b_start:c_start]

def possible_transposes(shape_len):
    for b_start in range(shape_len - 1):
        for c_start in range(b_start + 1, shape_len):
            yield b_start, c_start

def get_operations(source, target):
    visited = set([source])
    actions = list(possible_transposes(len(source)))

    def traverse(node, breadcrumbs, current_best):
        if current_best is not None and len(breadcrumbs) >= len(current_best):
            return current_best

        for b_start, c_start in actions:
            result = transpose_axes(node, b_start, c_start)
            if result in visited and result != target:
                continue
            visited.add(result)

            new_breadcrumbs = breadcrumbs + ((b_start, c_start),)

            if result == target:
                if current_best is None or len(current_best) > len(new_breadcrumbs):
                    return new_breadcrumbs

            current_best = traverse(result, new_breadcrumbs, current_best)
        return current_best

    return traverse(source, tuple(), None)

def get_transposes(shape, axes=None):

    source = tuple(range(len(axes)))
    if axes is None:
        axes = tuple(reversed(axes))
    else:
        assert set(source) == set(axes)

    for i in range(len(source) - 1, 0, -1):
        if source[:i] == axes[:i]:
            result = get_operations(source[i:], axes[i:])
            prefix = source[:i]
            break
    else:
        result = get_operations(source, axes)
        prefix = tuple()

    operations = [(b + len(prefix), c + len(prefix)) for b, c in result]

    transposes = []
    for b_start, c_start in operations:
        transposes.append((shape[:b_start], shape[b_start:c_start], shape[c_start:]))
        shape = transpose_axes(shape, b_start, c_start)
    return transposes


class Transpose(Computation):
    """
    Bases: :py:class:`~reikna.core.Computation`

    Changes the order of axes in a multidimensional array.
    Works analogous to ``numpy.transpose``.

    :param arr_t: an array-like defining the initial array.
    :param axes: tuple with the new axes order.
        If ``None``, then axes will be reversed.

    .. py:function:: compiled_signature(output:o, input:i)

        :param output: an array with all the attributes of ``arr_t``,
            with the shape permuted according to ``axes``.
        :param input: an array with all the attributes of ``arr_t``.
    """

    def __init__(self, arr_t, axes=None, block_width_override=None):

        self._block_width_override = block_width_override

        all_axes = range(len(arr_t.shape))
        if axes is None:
            axes = tuple(reversed(all_axes))
        else:
            assert set(axes) == set(all_axes)

        self._axes = tuple(axes)

        output_shape = transpose_shape(arr_t.shape, self._axes)
        output_arr = Type(arr_t.dtype, output_shape)

        Computation.__init__(self, [
            Parameter('output', Annotation(output_arr, 'o')),
            Parameter('input', Annotation(arr_t, 'i'))])

    def _add_transpose(self, plan, device_params,
            mem_out, mem_in, batch_shape, height_shape, width_shape):

        bso = self._block_width_override
        block_width = device_params.local_mem_banks if bso is None else bso

        if block_width ** 2 > device_params.max_work_group_size:
            # If it is not CPU, current solution may affect performance
            block_width = int(numpy.sqrt(device_params.max_work_group_size))

        input_height = helpers.product(height_shape)
        input_width = helpers.product(width_shape)
        batch = helpers.product(batch_shape)

        blocks_per_matrix = helpers.min_blocks(input_height, block_width)
        grid_width = helpers.min_blocks(input_width, block_width)

        render_kwds = dict(
            input_width=input_width, input_height=input_height, batch=batch,
            block_width=block_width,
            grid_width=grid_width,
            blocks_per_matrix=blocks_per_matrix,
            input_slices=[len(batch_shape), len(height_shape), len(width_shape)],
            output_slices=[len(batch_shape), len(width_shape), len(height_shape)])

        plan.kernel_call(
            TEMPLATE.get_def('transpose'), [mem_out, mem_in],
            global_size=(batch, blocks_per_matrix * block_width, grid_width * block_width),
            local_size=(1, block_width, block_width),
            render_kwds=render_kwds)

    def _build_plan(self, plan_factory, device_params, output, input_):
        plan = plan_factory()
        transposes = get_transposes(input_.shape, self._axes)

        mem_out = None
        for i, transpose in enumerate(transposes):

            batch_shape, height_shape, width_shape = transpose

            mem_in = input_ if i == 0 else mem_out
            if i == len(transposes) - 1:
                mem_out = output
            else:
                mem_out = plan.temp_array(
                    batch_shape + width_shape + height_shape, output.dtype)

            self._add_transpose(plan, device_params,
                mem_out, mem_in, batch_shape, height_shape, width_shape)

        return plan



"""
This module contains :py:class:`~reikna.cluda.Module` factories
which are used to compensate for the lack of complex number operations in OpenCL,
and the lack of C++ synthax which would allow one to write them.
"""

from warnings import warn

import numpy

from reikna.helpers import template_for
from reikna.cluda import dtypes
from reikna.cluda import Module


TEMPLATE = template_for(__file__)


def check_information_loss(out_dtype, expected_dtype):
    if dtypes.is_complex(expected_dtype) and not dtypes.is_complex(out_dtype):
        warn("Imaginary part ignored during the downcast from " +
            str(expected_dtype) + " to " + str(out_dtype),
            numpy.ComplexWarning)


def derive_out_dtype(out_dtype, *in_dtypes):
    expected_dtype = dtypes.result_type(*in_dtypes)
    if out_dtype is None:
        out_dtype = expected_dtype
    else:
        check_information_loss(out_dtype, expected_dtype)
    return out_dtype


def cast(out_dtype, in_dtype):
    """
    Returns a :py:class:`~reikna.cluda.Module` with a function of one argument
    that casts values of ``in_dtype`` to ``out_dtype``.
    """
    return Module(
        TEMPLATE.get_def('cast'),
        render_kwds=dict(out_dtype=out_dtype, in_dtype=in_dtype))


def add(*in_dtypes, **kwds):
    """add(*in_dtypes, out_dtype=None)

    Returns a :py:class:`~reikna.cluda.Module`  with a function of
    ``len(in_dtypes)`` arguments that adds values of types ``in_dtypes``.
    If ``out_dtype`` is given, it will be set as a return type for this function.

    This is necessary since on some platforms the ``+`` operator for a complex and a real number
    works in an unexpected way (returning ``(a.x + b, a.y + b)`` instead of ``(a.x + b, a.y)``).
    """
    assert set(kwds.keys()).issubset(['out_dtype'])
    out_dtype = derive_out_dtype(kwds.get('out_dtype', None), *in_dtypes)
    return Module(
        TEMPLATE.get_def('add_or_mul'),
        render_kwds=dict(op='add', out_dtype=out_dtype, in_dtypes=in_dtypes))


def mul(*in_dtypes, **kwds):
    """mul(*in_dtypes, out_dtype=None)

    Returns a :py:class:`~reikna.cluda.Module`  with a function of
    ``len(in_dtypes)`` arguments that multiplies values of types ``in_dtypes``.
    If ``out_dtype`` is given, it will be set as a return type for this function.
    """
    assert set(kwds.keys()).issubset(['out_dtype'])
    out_dtype = derive_out_dtype(kwds.get('out_dtype', None), *in_dtypes)
    return Module(
        TEMPLATE.get_def('add_or_mul'),
        render_kwds=dict(op='mul', out_dtype=out_dtype, in_dtypes=in_dtypes))


def div(in_dtype1, in_dtype2, out_dtype=None):
    """
    Returns a :py:class:`~reikna.cluda.Module` with a function of two arguments
    that divides values of ``in_dtype1`` and ``in_dtype2``.
    If ``out_dtype`` is given, it will be set as a return type for this function.
    """
    out_dtype = derive_out_dtype(out_dtype, in_dtype1, in_dtype2)
    return Module(
        TEMPLATE.get_def('div'),
        render_kwds=dict(out_dtype=out_dtype, in_dtype1=in_dtype1, in_dtype2=in_dtype2))


def conj(dtype):
    """
    Returns a :py:class:`~reikna.cluda.Module` with a function of one argument
    that conjugates the value of type ``dtype`` (must be a complex data type).
    """
    if not dtypes.is_complex(dtype):
        raise NotImplementedError("conj() of " + str(dtype) + " is not supported")

    return Module(
        TEMPLATE.get_def('conj'),
        render_kwds=dict(dtype=dtype))


def polar_unit(dtype):
    """
    Returns a :py:class:`~reikna.cluda.Module` with a function of one argument
    that returns a complex number ``(cos(theta), sin(theta))``
    for a value ``theta`` of type ``dtype`` (must be a real data type).
    """
    if not dtypes.is_real(dtype):
        raise NotImplementedError("polar_unit() of " + str(dtype) + " is not supported")

    return Module(
        TEMPLATE.get_def('polar_unit'),
        render_kwds=dict(dtype=dtype))


def norm(dtype):
    """
    Returns a :py:class:`~reikna.cluda.Module` with a function of one argument
    that returns the 2-norm of the value of type ``dtype``
    (product by the complex conjugate if the value is complex, square otherwise).
    """
    return Module(
        TEMPLATE.get_def('norm'),
        render_kwds=dict(dtype=dtype))


def exp(dtype):
    """
    Returns a :py:class:`~reikna.cluda.Module` with a function of one argument
    that exponentiates the value of type ``dtype``
    (must be a real or complex data type).
    """
    if dtypes.is_integer(dtype):
        raise NotImplementedError("exp() of " + str(dtype) + " is not supported")

    if dtypes.is_real(dtype):
        polar_unit_ = None
    else:
        polar_unit_ = polar_unit(dtypes.real_for(dtype))
    return Module(
        TEMPLATE.get_def('exp'),
        render_kwds=dict(dtype=dtype, polar_unit_=polar_unit_))


def polar(dtype):
    """
    Returns a :py:class:`~reikna.cluda.Module` with a function of two arguments
    that returns the complex-valued ``rho * exp(i * theta)``
    for values ``rho, theta`` of type ``dtype`` (must be a real data type).
    """
    if not dtypes.is_real(dtype):
        raise NotImplementedError("polar() of " + str(dtype) + " is not supported")

    return Module(
        TEMPLATE.get_def('polar'),
        render_kwds=dict(dtype=dtype, polar_unit_=polar_unit(dtype)))



"""
Test standard transformations
"""
import pytest

from reikna.cluda import Snippet
from reikna.algorithms import PureParallel
import reikna.transformations as tr
from reikna.core import Parameter, Annotation, Type

from helpers import *


def pytest_generate_tests(metafunc):
    int_dtypes = [numpy.dtype('int32'), numpy.dtype('int64')]
    real_dtypes = [numpy.dtype('float32')]
    complex_dtypes = [numpy.dtype('complex64')]

    if 'any_dtype' in metafunc.funcargnames:
        dtypes = int_dtypes + real_dtypes + complex_dtypes
        metafunc.parametrize('any_dtype', dtypes, ids=[str(x) for x in dtypes])

    if 'rc_dtype' in metafunc.funcargnames:
        dtypes = real_dtypes + complex_dtypes
        metafunc.parametrize('rc_dtype', dtypes, ids=[str(x) for x in dtypes])

    if 'dtype_to_broadcast' in metafunc.funcargnames:

        vals = []
        ids = []

        # a simple dtype
        vals.append(numpy.float32)
        ids.append('simple')

        # numpy itemsize == 9, but on device it will be aligned to 4, so the total size will be 12
        dtype = numpy.dtype([('val1', numpy.int32), ('val2', numpy.int32), ('pad', numpy.int8)])
        vals.append(dtype)
        ids.append("small_pad")

        dtype_nested = numpy.dtype([
            ('val1', numpy.int32), ('pad', numpy.int8)])
        dtype = numpy.dtype([
            ('val1', numpy.int32),
            ('val2', numpy.int16),
            ('nested', dtype_nested)])
        vals.append(dtype)
        ids.append("nested")

        dtype_nested = numpy.dtype(dict(
            names=['val1', 'pad'],
            formats=[numpy.int8, numpy.int8]))
        dtype = numpy.dtype(dict(
            names=['pad', 'struct_arr', 'regular_arr'],
            formats=[
                numpy.int32,
                numpy.dtype((dtype_nested, 2)),
                numpy.dtype((numpy.int16, (2, 3)))]))
        vals.append(dtype)
        ids.append("nested_array")

        metafunc.parametrize('dtype_to_broadcast', vals, ids=ids)


def get_test_computation(arr_t):
    return PureParallel(
        [Parameter('output', Annotation(arr_t, 'o')),
        Parameter('input', Annotation(arr_t, 'i'))],
        "${output.store_idx}(${idxs[0]}, ${input.load_idx}(${idxs[0]}));")


def test_copy(some_thr, any_dtype):

    input_ = get_test_array((1000,), any_dtype)
    input_dev = some_thr.to_device(input_)
    output_dev = some_thr.empty_like(input_dev)

    test = get_test_computation(input_dev)
    copy = tr.copy(input_dev)

    test.parameter.input.connect(copy, copy.output, input_prime=copy.input)
    test.parameter.output.connect(copy, copy.input, output_prime=copy.output)
    testc = test.compile(some_thr)

    testc(output_dev, input_dev)
    assert diff_is_negligible(output_dev.get(), input_)


def test_add_param(some_thr, any_dtype):

    input = get_test_array((1000,), any_dtype)
    p1 = get_test_array((1,), any_dtype)[0]
    p2 = get_test_array((1,), any_dtype)[0]
    input_dev = some_thr.to_device(input)
    output_dev = some_thr.empty_like(input_dev)

    test = get_test_computation(input_dev)
    add = tr.add_param(input_dev, any_dtype)

    test.parameter.input.connect(add, add.output, input_prime=add.input, p1=add.param)
    test.parameter.output.connect(add, add.input, output_prime=add.output, p2=add.param)
    testc = test.compile(some_thr)

    testc(output_dev, p1, input_dev, p2)
    assert diff_is_negligible(output_dev.get(), input + p1 + p2)


def test_add_const(some_thr, any_dtype):

    input = get_test_array((1000,), any_dtype)
    p1 = get_test_array((1,), any_dtype)[0]
    p2 = get_test_array((1,), any_dtype)[0]
    input_dev = some_thr.to_device(input)
    output_dev = some_thr.empty_like(input_dev)

    test = get_test_computation(input_dev)
    add1 = tr.add_const(input_dev, p1)
    add2 = tr.add_const(input_dev, p2)

    test.parameter.input.connect(add1, add1.output, input_prime=add1.input)
    test.parameter.output.connect(add2, add2.input, output_prime=add2.output)
    testc = test.compile(some_thr)

    testc(output_dev, input_dev)
    assert diff_is_negligible(output_dev.get(), input + p1 + p2)


def test_mul_param(some_thr, any_dtype):

    input = get_test_array((1000,), any_dtype)
    p1 = get_test_array((1,), any_dtype)[0]
    p2 = get_test_array((1,), any_dtype)[0]
    input_dev = some_thr.to_device(input)
    output_dev = some_thr.empty_like(input_dev)

    test = get_test_computation(input_dev)
    scale = tr.mul_param(input_dev, any_dtype)

    test.parameter.input.connect(scale, scale.output, input_prime=scale.input, p1=scale.param)
    test.parameter.output.connect(scale, scale.input, output_prime=scale.output, p2=scale.param)
    testc = test.compile(some_thr)

    testc(output_dev, p1, input_dev, p2)
    assert diff_is_negligible(output_dev.get(), input * p1 * p2)


def test_mul_const(some_thr, any_dtype):

    input = get_test_array((1000,), any_dtype)
    p1 = get_test_array((1,), any_dtype)[0]
    p2 = get_test_array((1,), any_dtype)[0]
    input_dev = some_thr.to_device(input)
    output_dev = some_thr.empty_like(input_dev)

    test = get_test_computation(input_dev)
    scale1 = tr.mul_const(input_dev, p1)
    scale2 = tr.mul_const(input_dev, p2)

    test.parameter.input.connect(scale1, scale1.output, input_prime=scale1.input)
    test.parameter.output.connect(scale2, scale2.input, output_prime=scale2.output)
    testc = test.compile(some_thr)

    testc(output_dev, input_dev)
    assert diff_is_negligible(output_dev.get(), input * p1 * p2)


def test_split_combine_complex(some_thr):

    i1 = get_test_array((1000,), numpy.float32)
    i2 = get_test_array((1000,), numpy.float32)
    i1_dev = some_thr.to_device(i1)
    i2_dev = some_thr.to_device(i2)
    o1_dev = some_thr.empty_like(i1)
    o2_dev = some_thr.empty_like(i2)

    base_t = Type(numpy.complex64, shape=1000)
    test = get_test_computation(base_t)
    combine = tr.combine_complex(base_t)
    split = tr.split_complex(base_t)

    test.parameter.input.connect(combine, combine.output, i_real=combine.real, i_imag=combine.imag)
    test.parameter.output.connect(split, split.input, o_real=split.real, o_imag=split.imag)
    testc = test.compile(some_thr)

    testc(o1_dev, o2_dev, i1_dev, i2_dev)
    assert diff_is_negligible(o1_dev.get(), i1)
    assert diff_is_negligible(o2_dev.get(), i2)


@pytest.mark.parametrize('order', [1, 2, 0.5])
def test_norm_param(some_thr, rc_dtype, order):

    input_ = get_test_array((1000,), rc_dtype)
    input_dev = some_thr.to_device(input_)

    norm = tr.norm_param(input_dev)

    output_dev = some_thr.empty_like(norm.output)

    test = get_test_computation(output_dev)
    test.parameter.input.connect(norm, norm.output, input_prime=norm.input, order=norm.order)
    testc = test.compile(some_thr)

    testc(output_dev, input_dev, order)
    assert diff_is_negligible(output_dev.get(), numpy.abs(input_) ** order)


@pytest.mark.parametrize('order', [1, 2, 0.5])
def test_norm_const(some_thr, rc_dtype, order):

    input_ = get_test_array((1000,), rc_dtype)
    input_dev = some_thr.to_device(input_)

    norm = tr.norm_const(input_dev, order)

    output_dev = some_thr.empty_like(norm.output)

    test = get_test_computation(output_dev)
    test.parameter.input.connect(norm, norm.output, input_prime=norm.input)
    testc = test.compile(some_thr)

    testc(output_dev, input_dev)
    assert diff_is_negligible(output_dev.get(), numpy.abs(input_) ** order)


def test_broadcast_const(some_thr, dtype_to_broadcast):

    dtype = dtypes.align(dtype_to_broadcast)
    const = get_test_array(1, dtype)[0]

    output_ref = numpy.empty((1000,), dtype)
    output_ref[:] = const

    output_dev = some_thr.empty_like(output_ref)

    test = get_test_computation(output_dev)
    bc = tr.broadcast_const(output_dev, const)
    test.parameter.input.connect(bc, bc.output)
    testc = test.compile(some_thr)

    testc(output_dev)
    assert diff_is_negligible(output_dev.get(), output_ref)


def test_broadcast_param(some_thr, dtype_to_broadcast):

    dtype = dtypes.align(dtype_to_broadcast)
    param = get_test_array(1, dtype)[0]

    output_ref = numpy.empty((1000,), dtype)
    output_ref[:] = param

    output_dev = some_thr.empty_like(output_ref)

    test = get_test_computation(output_dev)
    bc = tr.broadcast_param(output_dev)
    test.parameter.input.connect(bc, bc.output, param=bc.param)
    testc = test.compile(some_thr)

    testc(output_dev, param)
    assert diff_is_negligible(output_dev.get(), output_ref)
